import openai
import traceback
import asyncio
import sys
from dataclasses import dataclass
import random
import pickle
import os
import time
from functools import total_ordering

from .openai_api import complete, OpenAIRateLimitError, Capacity

class EmptyStreamError(Exception): pass

class ChaosException(openai.APIError): pass
class MaximumRetriesExceeded(Exception): 
    def __init__(self, error: Exception, retries: int):
        self.error = error
        self.retries = retries
    
    def __str__(self):
        print(self.error)
        return f"Maximum retries exceeded ({self.retries}) with error {type(self.error)}: {str(self.error)}"

class Batcher:
    def __init__(self, batch_size: int):
        self.tasks = []
        self.queued_requests = []
        self.batch_size = batch_size
    
    async def fill(self, queue: asyncio.Queue, maximum_collection_period: float = 0.1):
        if len(self.tasks) >= self.batch_size:
            return
        # first item is blocking call
        i = (await queue.get()).kwargs
        self.tasks.append(i)
        self.fill_nowait(queue)
        if len(self.tasks) <= self.batch_size:
            # wait some time if batch is not full yet
            await asyncio.sleep(maximum_collection_period)
            self.fill_nowait(queue)
        self.group()

    def fill_nowait(self, queue: asyncio.Queue):
        if queue.empty():
            pass
        else:
            try:
                while True and len(self.tasks) < self.batch_size:
                    self.tasks.append(queue.get_nowait().kwargs)
            except asyncio.QueueEmpty:
                pass
        
    def task_type(self, task):
        keys = ["model", "max_tokens", "temperature", "logprobs", "user", "logit_bias", "echo"]
        def get(k): 
            if k == "logit_bias": return "-".join([f"{k}={v}" for k,v in sorted(task.get(k, {}).items())])
            return str(task.get(k, "<none>"))
        identifier = "|".join([f"{k}={get(k)}" for k in keys])
        return identifier

    def group(self):
        assert len(self.queued_requests) == 0, f"Batcher.groups() called before self.queued_requests was emptied"

        buckets = {}
        
        for t in self.tasks:
            identifier = self.task_type(t)
            buckets.setdefault(identifier, []).append(t)
        
        for bucket in buckets.values():
            if "turbo" in bucket[0]["model"]:
                for t in bucket:
                    self.queued_requests.append(make_request_args([t]))
                continue
            self.queued_requests.append(make_request_args(bucket))
        
        self.tasks = []

def make_request_args(tasks):
    prompts = [t["prompt"] for t in tasks]
    futures = [t["future"] for t in tasks]
    request_ids = [t["request_id"] for t in tasks]

    # construct request arguments
    request_args = tasks[0].copy()
    del request_args["future"]
    
    request_args["prompt"] = prompts
    request_args["futures"] = futures
    request_args["request_id"] = request_ids
    request_args["stream"] = True

    return request_args

@dataclass
class Stats:
    prompt_tokens: int = 0
    tokens: int = 0
    requests: int = 0
    errors: int = 0
    sum_batch_size: int = 0

    def reset(self):
        self.tokens = 0
        self.requests = 0
        self.errors = 0
        self.sum_batch_size = 0

    def print(self):
        print(f"OpenAI API Stats: {self.requests} requests, {self.errors} errors, {self.tokens} tokens, {self.sum_batch_size} batch size, {float(self.sum_batch_size)/max(1,self.requests)} average batch size")
    
    def __str__(self):
        return f"OpenAI API Stats: {self.requests} requests, {self.errors} errors, {self.tokens} tokens, {float(self.sum_batch_size)/max(1,self.requests)} average batch size, reserved capacity {Capacity.reserved}/{Capacity.total}"

    def cost_estimate(self, model):
        k_tokens = float(self.tokens) / 1000

        # translate the above to python
        if model is None:
            print("warning: cost_estimate(): no model specified.")
            return -1
        if "text-davinci" in model:
            return k_tokens * 0.02
        elif "text-ada" in model:
            return k_tokens * 0.0004
        elif "text-babbage" in model:
            return k_tokens * 0.0005
        elif "text-curie" in model:
            return k_tokens * 0.002
        else:
            print("warning: cost_estimate(): unknown model {}".format(model))
            return -1

class ResponseStream:    
    def __init__(self, scheduler, kwargs, response, n, request_ids, maximum_retries=20, chaos = None, stats: Stats=None):
        self.scheduler: AsyncOpenAIAPI = scheduler
        self.kwargs = kwargs
        self.response = response
        self.request_ids = request_ids
        self.slices = [ResponseStreamSlice(self, self.view_kwargs(i), maximum_retries=maximum_retries) for i in range(n)]
        self.chaos = chaos
        self.stats = stats

        self.stats.requests += 1
        self.stats.sum_batch_size += n

        # task that always waits for new data in the response stream
        self.iteration_task = asyncio.create_task(self.iter_task())
    
    def view_kwargs(self, i):
        kwargs = self.kwargs.copy()
        kwargs["prompt"] = kwargs["prompt"][i]
        kwargs["request_id"] = self.request_ids[i]
        return kwargs

    def __del__(self):
        self.iteration_task.cancel()

    async def iter_task(self):
        try:
            self.response = aiter(self.response)
            async for data in self.response:
                if self.chaos is not None and random.random() > (1.0 - self.chaos):
                    raise ChaosException()
                
                if not "choices" in data.keys():
                    print("No choices in data", data)
                    continue

                for c in data["choices"]:
                    index = c["index"]

                    self.stats.tokens += len(c["logprobs"]["tokens"])
                    assert c is not None
                    self.slices[index].digest(c)
                    
                    if c["finish_reason"] == "stop":
                        # "<|endoftext|>", "<|endoftext|>", 0.0
                        self.slices[index].digest({
                            "text": "<|endoftext|>",
                            "logprobs": {
                                "text_offset": [0],
                                "token_logprobs": [0.0],
                                "tokens": ["<|endoftext|>"],
                                "top_logprobs": [{"<|endoftext|>": 0.0}]
                            }
                        })
                    
                    # logprobs.tokens, text, logprobs.token_logprobs
            for c in self.slices:
                c.finish()
        except Exception as e:
            print("Failed with", e)
            for c in self.slices:
                c.error(e)

    def view(self, index):
        assert index < len(self.slices), f"index {index} out of bounds for {len(self.slices)} slices of response stream"
        return self.slices[index]

@dataclass
class RecoveryAttempt:
    kwargs: dict
    error: Exception
    maximum_retries: int


class response_buffer_slice:
    def __init__(self, buffer, lower):
        self.buffer = buffer
        self.lower = lower
    
    def __str__(self) -> str:
        buffered_tokens = max(0, self.buffer.num_tokens - self.lower)
        return "<response_buffer_slice lower={} tokens_left≥{} >".format(self.lower, buffered_tokens)

    def __repr__(self) -> str:
        return str(self)

    async def empty(self):
        try:
            await self.get(0)
            return False
        except IndexError:
            return True

    async def get(self, i):
        return await self.buffer.get(i + self.lower)

    def __getitem__(self, i):
        if type(i) is slice:
            return response_buffer_slice(self.buffer, self.lower + i.start)
        assert False, f"response_buffer_slice.__getitem__({i}) not supported. Use async get() instead."

async def async_buffer(iterator, eager=False, tokenizer=None):
    if type(iterator) is list:
        # wrap already buffered data as response_buffer
        return response_buffer(None, iterator, tokenizer=tokenizer)

    if eager:
        data = []
        async for i in iterator: data.append(i)
        return response_buffer(None, data, tokenizer=tokenizer)
    else:
        if type(iterator) is ResponseStreamSlice:
            iterator = aiter(iterator)
        return response_buffer(iterator, tokenizer=tokenizer)
class response_buffer:
    def __init__(self, iterator, fixed_data=None, tokenizer=None):
        self.iterator = iterator

        self.text = ""
        self.num_tokens = 0
        self.logprobs = {
            "text_offset": [],
            "token_logprobs": [],
            "tokens": [],
            "top_logprobs": []
        }

        if fixed_data is not None:
            self.fixed = True
            self._append(fixed_data)
            self.tokenizer = None
        else:
            self.fixed = False
            # when provided, convert ["logprobs"]["tokens"] to token IDs automatically
            self.tokenizer = tokenizer
            assert self.tokenizer is not None, f"response_buffer: tokenizer must be provided when using non-fixed data"
            

    def __str__(self) -> str:
        return "<response_buffer num_tokens={} iterator={}>".format(self.num_tokens, self.iterator)
    
    @classmethod
    def singleton(cls, text=None, text_offset=None, token_logprob=None, token=None, top_logprobs=None):
        return cls(None, {
            "text": text or "",
            "logprobs": {
                "text_offset": [text_offset],
                "token_logprobs": [token_logprob],
                "tokens": [token],
                "top_logprobs": [top_logprobs]
            }
        })


    def _append(self, data):
        self.text += data["text"]
        self.logprobs["text_offset"] += data["logprobs"]["text_offset"]
        self.logprobs["token_logprobs"] += data["logprobs"]["token_logprobs"]
        self.logprobs["tokens"] += data["logprobs"]["tokens"]
        self.logprobs["top_logprobs"] += data["logprobs"]["top_logprobs"]
        
        self.num_tokens = len(self.logprobs["tokens"])

    # allow async iteration over response buffer
    def __aiter__(self):
        async def _aiter():
            i = 0
            while True:
                try:
                    yield await self.get(i)
                    i += 1
                except IndexError:
                    break
        return _aiter()

    async def get(self, i):
        while self.num_tokens <= i and self.iterator is not None:
            try:
                chunk = await anext(self.iterator)
                chunk["logprobs"]["tokens"] = await self.tokenizer(chunk["logprobs"]["tokens"])
                # self.tokenizer(data["logprobs"]["tokens"])
                self._append(chunk)
            except StopAsyncIteration:
                break
        if i >= self.num_tokens:
            raise IndexError(f"index {i} out of bounds for response_buffer of length {self.num_tokens}. Iterator is {self.iterator}")

        text_start = self.logprobs["text_offset"][i]
        text_end = self.logprobs["text_offset"][i+1] if i+1 < len(self.logprobs["text_offset"]) else None
        
        return {
            "text": self.text[text_start:text_end],
            "logprobs": {
                "text_offset": self.logprobs["text_offset"][i],
                "token_logprobs": self.logprobs["token_logprobs"][i],
                "tokens": self.logprobs["tokens"][i],
                "top_logprobs": self.logprobs["top_logprobs"][i]
            },
            **({"fixed": True} if self.fixed else {})
        }
    
    async def empty(self):
        try:
            await self.get(0)
            return False
        except IndexError:
            return True

    def __getitem__(self, i):
        # slice 
        if isinstance(i, slice):
            assert i.stop is None, "slicing with stop index not supported on OpenAIResponseBuffer"
            assert i.step is None, "slicing with step not supported on OpenAIResponseBuffer"
            return response_buffer_slice(self, i.start)
        else:
            assert False, "only slicing supported on response_buffer. For single item access, use async get()"

class ResponseStreamSliceIterator:
    def __init__(self, slice):
        self.slice = slice
        self.retries = 0
        self.text = ""
        self.consumed_tokens = []

    async def recover(self):
        recovery_kwargs = self.slice.kwargs.copy()
        # reconstruct the prompt by tokenizing the consumed tokens
        if len(self.consumed_tokens) > 0:
            prompt = self.consumed_tokens
            recovery_kwargs["prompt"] = [t[0] for t in prompt]
        
        # issue new completion call
        new_slice = await self.slice.stream.scheduler.complete(**recovery_kwargs)
        new_it = ResponseStreamSliceIterator(new_slice)
        new_it.retries = self.retries + 1

        # print("recovery for request with ID", recovery_kwargs["request_id"])
        
        # skip as many data packets as necessary to get to the original point of failure
        while len(new_it.consumed_tokens) < len(self.consumed_tokens):
            last_data = await anext(new_it)
            
            # if last chunk of new stream is too long, we return a partial chunk to align
            if len(new_it.consumed_tokens) > len(self.consumed_tokens):
                offset = len(new_it.consumed_tokens) - len(self.consumed_tokens)
                partial_data = {
                    "text": new_it.text[len(self.text):],
                    "logprobs": {
                        "text_offset": last_data["logprobs"]["text_offset"][-offset:],
                        "token_logprobs": last_data["logprobs"]["token_logprobs"][-offset:],
                        "tokens": last_data["logprobs"]["tokens"][-offset:],
                        "top_logprobs": last_data["logprobs"]["top_logprobs"][-offset:]
                    }
                }

                self.text = new_it.text
                self.consumed_tokens = new_it.consumed_tokens
                self.slice = new_slice
                self.retries = new_it.retries
                
                return partial_data
        self.text = new_it.text
        self.consumed_tokens = new_it.consumed_tokens
        self.slice = new_slice
        # otherwise the chunking aligns with the old stream, so we return the next chunk
        return await self.__anext__()

    async def __anext__(self):
        try:
            try:
                data = await asyncio.wait_for(self.slice.data_queue.get(), 2.0)
            except asyncio.TimeoutError as e:
                data = RecoveryAttempt(self.slice.kwargs, e, self.slice.maximum_retries)
            # None indicates end of stream
            if data is None: raise StopAsyncIteration
            # exceptions that are queued are definitive (all retries failed)
            if isinstance(data, Exception): raise data
            # RecoveryAttempt indicates that the underlying stream errored out and we need to recover (still retries left)
            if isinstance(data, RecoveryAttempt):
                # if the stream of our self.slice errors out, we can recover by creating a new 
                # stream via a new call to openai.Completion.create
                attempt: RecoveryAttempt = data
                traceback.print_exc()
                print("OpenAI API: Underlying stream of OpenAI complete() call failed with error", type(attempt.error), attempt.error, f"Retrying... (attempt: {self.retries})", flush=True)
                self.retries += 1
                # if we have exceeded the maximum number of retries, raise the error
                if self.retries > attempt.maximum_retries:
                    raise MaximumRetriesExceeded(attempt.error, retries=self.retries)
                if self.slice.stream.scheduler.tokenizer is None:
                    print("Cannot recover from stream error without a configured tokenizer", flush=True)
                    raise attempt.error
                return await self.recover()
            
            self.consumed_tokens += data["logprobs"]["tokens"]
            self.text += data["text"]

            return data
        except asyncio.CancelledError:
            raise StopAsyncIteration


class ResponseStreamSlice:
    def __init__(self, stream, kwargs, maximum_retries=3):
        self.stream: ResponseStream = stream
        self.kwargs = kwargs
        self.maximum_retries = maximum_retries

        self.data_queue = asyncio.Queue()
        self.failed = False

    def digest(self, data):
        assert not self.failed, f"digest called on failed slice"
        self.data_queue.put_nowait(data)
    
    def finish(self):
        assert not self.failed, f"finish called on failed slice"
        self.data_queue.put_nowait(None)

    def error(self, error):
        assert not self.failed, f"error called on failed slice"
        self.failed = True
        self.data_queue.put_nowait(RecoveryAttempt(self.kwargs, error, self.maximum_retries))

    def __aiter__(self):
        return ResponseStreamSliceIterator(self)


@dataclass
@total_ordering
class RequestQueueItem:
    kwargs: dict
    priority: int

    # comparison
    def __lt__(self, other):
        return self.priority < other.priority
    
    def __eq__(self, other):
        return self.priority == other.priority

class AsyncOpenAIAPI:
    def __init__(self):
        self.maximum_retries = 20
        
        self.complete_api_call_queue = asyncio.PriorityQueue()
        self.complete_api_worker = asyncio.create_task(self.api_complete_worker(self.complete_api_call_queue))
        
        self.request_ctr = 0
        self.request_ctr_offset = 1000000000
        
        self.complete_request_queue = asyncio.Queue()
        self.complete_request_workers = [asyncio.create_task(self.complete_request_worker(self.complete_request_queue)) for i in range(5)]

        self.stats_logger = None 
        
        # chaos debugging (introduces random failures in the OpenAI API)
        self.chaos = None
        self.warned_about_chaos = False
        
        self.batch_size = 20
        self.maximum_collection_period = 0.05

        self.stats = Stats()
        self.nostream = False
        
        # INTERNAL OPTION only. In theory we can do caching but there are consequences
        # deterministic sampling, large cache size, cache loading startup time, etc.
        # also, when exposed to clients, this should be implemented on a per query level, not per batches
        self.use_cache = False

        self.tokenizer = None
        
        self.cache = {}
        self.cache_dir = "."
        self.futures = set()
        self.restore_cache()

    def restore_cache(self):
        if not self.use_cache:
            return
        cache_file = "openai.completions.cache"
        if os.path.exists(os.path.join(self.cache_dir, cache_file)):
            with open(os.path.join(self.cache_dir, cache_file), "rb") as f:
                self.cache = pickle.load(f)
    
    def save_cache(self):
        cache_file = "openai.completions.cache"
        with open(os.path.join(self.cache_dir, cache_file), "wb") as f:
            pickle.dump(self.cache, f)

    def start_stats_logger(self):
        self.stats_logger = asyncio.create_task(self.stats_logger_worker())
    def stop_stats_logger(self):
        self.stats_logger.cancel()

    async def stats_logger_worker(self):
        while True:
            await asyncio.sleep(1)
            print(self.stats, flush=True)

    def warn_chaos(self):
        if self.chaos is not None:
            if self.warned_about_chaos: return
            print("warning: AsyncOpenAIAPI.set_chaos() is set to a value different from None. This is only for testing purposes and should not be used in production (makes OpenAI complete streams fail randomly on purpose).")
            self.warned_about_chaos = True

    def set_chaos(self, chaos):
        self.chaos = chaos
        self.warn_chaos()

    def __del__(self):
        if self.stats_logger is not None:
            self.stats_logger.cancel()
        # cancel the score worker task
        self.complete_api_worker.cancel()
        for worker in self.complete_request_workers:
            worker.cancel()
        try:
            loop = asyncio.get_event_loop()
            while not all([t.done() for t in (self.complete_request_workers + [self.complete_api_worker])]):
                loop._run_once()
        except:
            pass # if no more event loop is around, no need to wait for the workers to finish

    async def api_complete_worker(self, queue):
        while True:
            self.futures = set([f for f in self.futures if not f.done()])
            while Capacity.reserved >= Capacity.total * 0.8:
                # print("wait before queing more requests", flush=True)
                await asyncio.sleep(0.1)
                # print(Capacity.reserved, Capacity.total, flush=True)
            # print(Capacity.reserved, Capacity.total, flush=True)
            batcher = Batcher(self.batch_size)
            await batcher.fill(queue, maximum_collection_period=self.maximum_collection_period)
            for kwargs in batcher.queued_requests:
                await self.complete_request_queue.put(kwargs)

    async def _create(self, **kwargs):
        async def first_buffered(aiter, first):
            yield first
            async for x in aiter:
                yield x

        num_prompt_tokens = sum([len(p) for p in kwargs["prompt"]])
        self.stats.prompt_tokens += num_prompt_tokens

        res = complete(**kwargs)
        first = await anext(res)
        return first_buffered(res, first)
        
        return res

    def is_definitive_error(self, e):
        if "logit biases, but can provide at most" in str(e):
            return True
        return False

    async def complete_request_worker(self, queue: asyncio.Queue):
        while True:
            try:
                kwargs = await queue.get()
                futures = kwargs.pop("futures")
                request_ids = kwargs.pop("request_id")
                retries = self.maximum_retries
                while True:
                    try:
                        if retries != self.maximum_retries:
                            print("Retrying", retries, "more times")
                            await asyncio.sleep(0.5)
                        res = await self._create(**kwargs)
                        break
                    except Exception as e:
                        if type(e) is AssertionError:
                            raise e
                        self.stats.errors += 1
                        retries -= 1            
                        print("Failed to call complete endpoint", type(e), e, flush=True)
                        await asyncio.sleep(0.5)
                        if retries <= 0 or self.is_definitive_error(e):
                            raise e
                        if type(e) is TimeoutError or type(e) is OpenAIRateLimitError:
                            t = (2.0 * random.random()) ** (self.maximum_retries - retries)
                            print("Backing off for", t , "seconds")
                            await asyncio.sleep(t)
            except asyncio.CancelledError:
                return
            except Exception as e:
                print("error", type(e))
                for future in futures:
                    future.set_exception(e)
                continue

            self.warn_chaos() # warns about self.chaos if set

            rsi = ResponseStream(self, kwargs, res, len(futures), maximum_retries=self.maximum_retries, chaos=self.chaos, stats=self.stats, request_ids = request_ids)
            for i, future in enumerate(futures):
                future.set_result(rsi.view(i))

    async def complete(self, request_id=None, **kwargs):
        assert "prompt" in kwargs, f"bopenai requires prompt to be set"

        loop = asyncio.get_running_loop()
        result_fut = loop.create_future()
        self.futures.add(result_fut)

        if request_id is None:
            request_id = self.request_ctr
            self.request_ctr += 1
        else:
            print("re-trying request id", request_id)
        
        kwargs = {"future": result_fut, "request_id": request_id, **kwargs}
        
        if "logit_bias" in kwargs and len(kwargs["logit_bias"]) > 300:
            biases = list(kwargs["logit_bias"].items())
            # make sure to always include eos if set and truncating
            if 50256 in kwargs["logit_bias"]:
                biases = biases[:299] + [(50256, kwargs["logit_bias"][50256])]
            else:
                biases = biases[:300]
            print("warning: the required logit_bias is too large to be handled by the OpenAI API and will be limited to the first 300 tokens. This can lead to the violation of the provided constraints or undesired model output. To avoid this use less broad or no constraints.")
            kwargs["logit_bias"] = {t:b for t,b in biases}


        assert kwargs.get("echo", False), f"bopenai requires echo=True for to enable proper error recovery. Please handle proper prompt removal in client code."

        r = RequestQueueItem(kwargs, request_id)
        await self.complete_api_call_queue.put(r)
        self.request_ctr += 1
        
        return await result_fut


"""
Implementation of on-the-fly input_id rewriting 
during decoding.
"""

from typing import Optional, Dict, Any, Union

from dataclasses import dataclass
from typing import List
import asyncio

import numpy as np
from lmql.utils import nputil

from lmql.runtime.decoder_head import DecoderHead


@dataclass
class ActivePromptTokens:
    next_tokens: np.ndarray # long
    is_actively_prompted: np.ndarray # bool

def dim(t):
    if type(t) is int:
        return 0
    elif type(t) is list:
        if len(t) == 0:
            return 2
        return 1 + dim(t[0])
    elif nputil.is_array(t):
        return t.ndim
    else:
        raise Exception("ndim not implemented for type: " + str(type(t)))

@dataclass
class RewrittenInputIds:
    appended_input_ids: List[np.ndarray]
    strip_eos: bool = True
    user_data: Optional[Union[List[Dict[str, Any]], Dict[str, Any]]] = None

    def strip_ids(self, input_ids, i):
        if self.strip_eos == True:
            return input_ids[i,:-1]
        elif type(self.strip_eos) is list:
            sequence_strip = self.strip_eos[i]
            if sequence_strip == True:
                return input_ids[i,:-1]
            elif type(sequence_strip) is int:
                return input_ids[i,:sequence_strip]
            else:
                return input_ids[i]
        else:
            return input_ids[i]

    def setting_user_data(self, key, value):
        if self.user_data is None:
            self.user_data = {}
        self.user_data[key] = value
        return self

    def append(self, input_ids: np.ndarray):
        to_append = self.appended_input_ids
        if to_append is None: to_append = [np.array([], dtype=np.int64, device=input_ids.device) for _ in range(len(input_ids))]

        # concat existing input_ids (minus eos if self.strip_eos) with appended input_ids (head_seqs)
        input_ids = [np.concatenate([self.strip_ids(input_ids, i), ids], axis=-1) for i, ids in enumerate(to_append)]

        return input_ids

    def append_and_left_pad(self, input_ids: np.ndarray):
        device =  input_ids.device
        input_ids = self.append(input_ids)
        max_length = max(len(seq) for seq in input_ids)
                
        # padding masks
        seqs_padding = [max_length - len(seq) for seq in input_ids]
        seqs_padding = np.stack([np.arange(max_length) < padding for padding in seqs_padding], axis=0)
        # pad actual input IDs
        input_ids = np.stack([left_pad(seq, max_length) for seq in input_ids], axis=0)
        
        assert input_ids.dtype == np.int64
        
        return input_ids, seqs_padding

    @staticmethod
    def with_user_data(rewritten_ids, key, value):
        if rewritten_ids is None: 
            rewritten_ids = RewrittenInputIds(appended_input_ids=None, strip_eos=False)
        return rewritten_ids.setting_user_data(key, value)

def left_pad(t, n, value = -1):
    if len(t) >= n: return t
    return np.concatenate([np.ones(n - len(t), dtype=np.int64) * value, t], axis=0)

class InputIdRewriter:
    def __init__(self, head_rewriter, tokenize, detokenize, eos_token_id, initial_prompt_offset):
        self.head_rewriter = head_rewriter
        
        self.tokenize = tokenize
        self.detokenize = detokenize

        self.eos_token_id = eos_token_id
        self.initial_prompt_offset = initial_prompt_offset

    async def rewrite_seq(self, i: int, head_input_ids, next_token_scores, next_token_logprob: Optional[np.ndarray], mask_seq_to_rewrite: Optional[np.ndarray] = None, user_data=None):
        # only rewrite masked sequences if a mask is provided
        if mask_seq_to_rewrite is not None:
            if not mask_seq_to_rewrite[i]:
                return None
        
        head = DecoderHead(i, 
            head_input_ids[:-1], head_input_ids[-1], next_token_scores[i],
            self.tokenize,
            self.detokenize,
            self.eos_token_id,
            self.initial_prompt_offset,
            next_token_logprob = next_token_logprob[i] if next_token_logprob is not None else None,
            user_data = user_data
        )
        return await self.head_rewriter(head)

    async def input_ids_rewriter_fn(self, input_ids, next_token_scores, seq_idx = None, next_token_logprob = None, mask_seq_to_rewrite = None, user_data=None):
        def get_user_data(i):
            if user_data is None: 
                return None
            return user_data[i]
        
        if dim(input_ids) == 2:
            head_results: List[RewrittenInputIds] = await asyncio.gather(*(self.rewrite_seq(i, head_input_ids, next_token_scores, next_token_logprob, mask_seq_to_rewrite, user_data=get_user_data(i)) for i, head_input_ids in enumerate(input_ids)))

            if all(r is None for r in head_results):
                # note: user data will be lost here
                return RewrittenInputIds(appended_input_ids=None, strip_eos=False, user_data=[None for _ in range(len(input_ids))])
            else:
                strip_eos = [r.strip_eos if r is not None else False for r in head_results]
                if all([s == True for s in strip_eos]): strip_eos = True

                # collect appended input ids
                appended_seqs = [rewrite.appended_input_ids if rewrite is not None else [] for rewrite in head_results]
                appended_seqs = [nputil.ensure_array(seq) for seq in appended_seqs]

                # collect user data
                user_data = [rewrite.user_data if rewrite is not None else None for rewrite in head_results]
                
                # handle case where nothing needs to be appended but we still need to strip eos
                if all(seq is None for seq in appended_seqs):
                    appended_seqs = None

                return RewrittenInputIds(appended_input_ids=appended_seqs, strip_eos=strip_eos, user_data=user_data)
        else: # must be dim() == 1
            assert False, "warning: 1D input_ids are deprecated and will be removed in a future version."
            assert input_ids.dim() == 1, "input_ids must be 1D or 2D for InputIdRewriter"
            assert seq_idx is not None, "seq_idx must be provided for 1D input_ids"
            assert mask_seq_to_rewrite is None, "mask_seq_to_rewrite is not supported for 1D input_ids"

            result = await self.head_rewriter(DecoderHead(seq_idx, input_ids, None, self.tokenize, self.detokenize, self.eos_token_id, self.initial_prompt_offset))

            if result is None:
                return RewrittenInputIds(appended_input_ids=None, strip_eos=False)
            else:
                return RewrittenInputIds(appended_input_ids=ensure_tensor(result.appended_input_ids, input_ids.device), strip_eos=result.strip_eos)

class ActivePromptingRewriter:
    def __init__(self, head_rewriter, tokenize, detokenize, eos_token_id, initial_prompt_offset):
        self.head_rewriter = head_rewriter
        
        self.tokenize = tokenize
        self.detokenize = detokenize

        self.eos_token_id = eos_token_id
        self.initial_prompt_offset = initial_prompt_offset

    async def input_ids_rewriter_fn(self, input_ids, active: bool):
        assert active, "ActivePromptingRewriter should only be used when actively prompting (active=True)"
        # print("ActivePromptingRewriter: input_ids", len(input_ids))

        head_results: List[ActivePromptTokens] = [
            await self.head_rewriter(DecoderHead(i, 
                head_input_ids[:-1], head_input_ids[-1], None,
                self.tokenize,
                self.detokenize,
                self.eos_token_id,
                self.initial_prompt_offset))
        for i, head_input_ids in enumerate(input_ids)]

        if all(r is None for r in head_results):
            all_zeros = np.zeros(input_ids.shape[0], device=input_ids.device, dtype=np.int64)
            return ActivePromptTokens(next_tokens=all_zeros, is_actively_prompted=all_zeros.clone())
        else:
            next_tokens = np.array([next_token if next_token is not None else -1 for next_token in head_results], device=input_ids.device, dtype=np.int64)
            is_actively_prompted = np.array([1 if r is not None else 0 for r in head_results], device=input_ids.device, dtype=np.int64)
            return ActivePromptTokens(next_tokens=next_tokens, is_actively_prompted=is_actively_prompted)


class ProgramState:
    def __init__(self, runtime=None):
        self.variable_values = {}
        self.variable_diffs = {}
        self.variable_scores = {}
        self.variable_monotonicity = {}

        self.runtime = runtime

    async def json(self):
        def json_value(k):
            fin = self.variable_monotonicity.get(k, "var")
            # do not include diff and score for now
            return fin + "(\"" + str(self.variable_values[k]) + "\")"
        return {k: json_value(k) for k in self.variable_values.keys()}

    def __getitem__(self, key):
        return self.get(key)

    def __setitem__(self, key, value):
        self.set(key, value)

    def set(self, name, value, scores=None, diff=None, montonicity="var"):
        self.variable_values[name] = value
        self.variable_diffs[name] = diff
        self.variable_scores[name] = scores
        self.variable_monotonicity[name] = montonicity

    def get(self, name, default=None):
        return self.variable_values.get(name, default)
    
    def get_diff(self, name, default=None):
        return self.variable_diffs.get(name, default)

    def final(self, name):
        return self.variable_monotonicity.get(name, "var")

    def copy(self):
        s = ProgramState()
        s.variable_values = self.variable_values.copy()
        s.variable_monotonicity = self.variable_monotonicity.copy()
        s.variable_diffs = self.variable_diffs.copy()
        s.variable_scores = self.variable_scores.copy()
        s.runtime = self.runtime
        return s


import re

from tokenize import Token
from typing import Iterable, Tuple
from itertools import product

from lmql.utils import nputil
import numpy as np

def get_vocab(tokenizer):
    if hasattr(tokenizer, "vocab"):
        return tokenizer.vocab
    elif hasattr(tokenizer, "get_vocab"):
        return tokenizer.get_vocab()
    else:
        assert False, "Could not obtain full vocabulary from unknown tokenizer type: {}".format(type(tokenizer))

class VocabularyMatcher:
    """
    Generates sub-token level logit masks from provided tokens.
    """
    def __init__(self, tokenizer):
        self.tokenizer = tokenizer
        self.vocab = {v: k for k, v in get_vocab(self.tokenizer).items()}

        # TODO: this should be more complete
        self.space_repr = self.tokenizer.tokenize(" ")[0]
        self.nl_repr = self.tokenizer.tokenize("\n")[0]
        
        self.token_lengths = None

    @property
    def eos_token_id(self):
        return self.tokenizer.eos_token_id

    @staticmethod
    def init(tokenizer):
        vm = VocabularyMatcher(tokenizer)
        VocabularyMatcher._instance = vm

    @staticmethod
    def instance():
        if VocabularyMatcher._instance is None:
            raise Exception("VocabularyMatcher not initialized.")
        return VocabularyMatcher._instance

    @staticmethod
    def ensure_ready():
        VocabularyMatcher.instance()

    @property
    def vocab_size(self):
        return len(self.vocab)

    def make_mask(self, tokens=None, regex=None, minus=None, prefix=False, exact=False, charlen=None):
        if tokens is not None:
            mask = self._make_mask_from_tokens(tokens, prefix, exact=exact)
        elif charlen is not None:
            mask = self._make_mask_from_char_length(charlen)
        else:
            assert regex is not None, "TokenSetConcrete: either tokens or regex must be set."
            assert not prefix, "TokenSetConcrete: prefix is not supported for regex."
            mask = self._make_mask_from_regex(regex)

        if minus: mask = np.logical_not(mask)

        return mask

    def _make_mask_from_regex(self, regex):
        regex = regex.replace(" ", self.space_repr)
        regex = regex.replace("\n", self.nl_repr)

        regex = regex.replace(" ", self.tokenizer.tokenize(" ")[0])

        mask = np.zeros([self.vocab_size], dtype=np.bool_)

        pattern = re.compile(regex, re.UNICODE)
        for id, subtoken in self.vocab.items():
            if pattern.match(subtoken) is not None:
                mask[id] = True

        return mask

    def _make_mask_from_char_length(self, length):
        if self.token_lengths is None:
            token_lengths = np.zeros([self.vocab_size], dtype=np.int32)
            for id, subtoken in self.vocab.items():
                token_lengths[id] = len(subtoken)
            self.token_lengths = token_lengths
        
        return self.token_lengths == length

    def _make_mask_from_tokens(self, tokens, prefix, exact=False):
        mask = np.zeros([self.vocab_size], dtype=np.bool_)

        if "*" in tokens:
            mask[:] = True
        elif len(tokens) > 0:
            if prefix:
                # instead of using the tokens themselves, use subtoken prefixes of the tokens
                tokens = [self.tokenizer(t)["input_ids"][0] for t in tokens]
                for t in tokens: 
                    mask[t] = True
            else:
                if any(t for t in tokens if t != "eos") > 0:
                    def process(t):
                        t = t.replace(".", "\\.")
                        t = t.replace(" ", self.space_repr)
                        t = t.replace("\n", self.nl_repr)
                        return t
                    if exact: 
                        # only allow exact matches
                        pattern = "|".join(f"({process(t)})" for t in tokens if t != "eos")
                        pattern = re.compile(pattern, re.UNICODE)
                        matcher = pattern.fullmatch
                    else:
                        # allow arbitrary further text per token
                        pattern = "|".join(f"{process(t)}.*" for t in tokens if t != "eos")
                        pattern = re.compile(pattern, re.UNICODE)
                        matcher = pattern.match
                    
                    for id, subtoken in self.vocab.items():
                        if matcher(subtoken) is not None:
                            mask[id] = True
                
                if any([t == "eos" for t in tokens]): # has eos
                    mask[self.eos_token_id] = True

        return mask

    def str(self, mask, full=False):
        prefix = ""
        tokens = []
        mask = mask

        if mask.sum() == mask.shape[0]:
            return "*"

        if mask.sum() > np.logical_not(mask).sum() and np.logical_not(mask).sum() > 0:
            prefix = "* \ "
            mask = np.logical_not(mask)

        truncated = False

        # get list of all non-zero indices in self.mask tensor
        for i in mask.reshape(-1).nonzero()[0]:
            if len(tokens) > 5 and not full:
                truncated = True
                break
            if i == self.eos_token_id:
                tokens.append("eos")
            else:
                s = self.vocab[i]
                # replace nl and space
                s = self.tokenizer.convert_tokens_to_string([s])
                s = s.encode("unicode_escape").decode("utf-8")
                tokens.append(s)

        return prefix + "{{{}}}".format(
            ", ".join([t for t in sorted(list(tokens))]) + ("..." if truncated else "")
        )

VocabularyMatcher._instance = None

class TokenSetConcrete:
    def __new__(cls, *args, **kwargs):
        if "mask" in kwargs.keys():
            mask = kwargs["mask"]
            return super(TokenSetConcrete, cls).__new__(cls)
        elif "regex" in kwargs.keys():
            regex = kwargs["regex"]
            key = "regex:" + regex
            if key in TokenSetConcrete.cache.keys():
                return TokenSetConcrete.cache[key]
            else:
                TokenSetConcrete.cache[key] = super(TokenSetConcrete, cls).__new__(cls)
                return TokenSetConcrete.cache[key]
        elif "charlen" in kwargs.keys():
            key = "charlen:" + str(kwargs["charlen"])
            if key in TokenSetConcrete.cache.keys():
                return TokenSetConcrete.cache[key]
            else:
                TokenSetConcrete.cache[key] = super(TokenSetConcrete, cls).__new__(cls)
                return TokenSetConcrete.cache[key]
        else:
            tokens = kwargs.get("tokens", args[0])
            assert tokens is not None
        
            minus = kwargs.get("minus", False)
            prefix = kwargs.get("prefix", False)
            
            t = ("prefix " if prefix else "") + ("* \ " if minus else "") + "|".join(sorted(list(tokens)))
            
            if t in TokenSetConcrete.cache.keys():
                return TokenSetConcrete.cache[t]
            else:
                TokenSetConcrete.cache[t] = super(TokenSetConcrete, cls).__new__(cls)
                return TokenSetConcrete.cache[t]

    def __init__(self, tokens=None, minus=False, mask=None, regex=None, prefix=False, exact=False, charlen=None):
        VocabularyMatcher.ensure_ready()

        if mask is not None:
            self.mask = mask.copy()
        else: 
            self.mask = VocabularyMatcher.instance().make_mask(tokens=tokens, regex=regex, minus=minus, prefix=prefix, exact=exact, charlen=charlen)

        self._token_str = None
        # for TokenSetSymbolic compatibility
        self.minusset = False

    def union(self, other):
        if other == "∅": 
            return TokenSetConcrete(mask=self.mask)
        if other == "*": 
            return "*"

        assert type(other) is TokenSetConcrete, "Can only union over two TokenSetConcrete."

        return TokenSetConcrete(mask=np.logical_or(self.mask, other.mask))

    def intersect(self, other):
        if other == "∅": return "∅"
        if other == "*": return self

        assert type(other) is TokenSetConcrete, "Can only intersect two TokenSetConcrete."

        return TokenSetConcrete(mask=np.logical_and(self.mask, other.mask))
     
    def setminus(self, other):
        if other == "*": 
            return "∅"
        if other == "∅":
            return TokenSetConcrete(mask=self.mask)
        
        assert type(other) is TokenSetConcrete, "Can only setminus two TokenSetConcrete."

        return TokenSetConcrete(mask=np.logical_and(self.mask, np.logical_not(other.mask)))
    

    def starts_with(self, s):
        if s in self.tokens: 
            return not self.minusset # returns True if not minusset and s in self.tokens
        else:
            for s in self.tokens:
                if s.startswith(s): 
                    return not self.minusset # returns True if not minusset and some s in self.tokens starts with s
            return self.minusset

    def __len__(self):
        return self.mask.sum()

    def __repr__(self):
        return str(self)

    def __str__(self, full=False):
        if self._token_str is not None:
            return self._token_str

        self._token_str = VocabularyMatcher.instance().str(self.mask, full=full)
        
        return self._token_str

    def __eq__(self, other):
        if other == "∅": 
            if self.mask.sum() == 0: 
                return True
            return False
        if other == "*": 
            if self.mask.sum() == self.mask.shape[0]: 
                return True
            return False

        assert type(other) is TokenSetConcrete, "Can only compare (==) two TokenSets."

        return np.all(self.mask == other.mask)
TokenSetConcrete.cache = {}

class TokenSetSymbolic:
    def __init__(self, tokens=None, minus=False):
        assert token_set_vocabulary is not None, "TokenSetConcrete: token_set_vocabulary must be set before any TokenSets are instantiated."

        if tokens is None: tokens = set()

        self.tokens = tokens
        self.minusset = minus

    def union(self, other):
        if other == "∅": 
            return TokenSetSymbolic(tokens=self.tokens, minus=self.minusset)
        if other == "*": 
            return "*"

        assert type(other) is TokenSetSymbolic, "Can only union over two TokenSetSymbolics."

        if self.minusset:
            if other.minusset:
                return TokenSetSymbolic(tokens=self.tokens.intersection(other.tokens), minus=True)
            else:
               return TokenSetSymbolic(tokens=self.tokens - other.tokens, minus=True) 
        else:
            if other.minusset:
                return TokenSetSymbolic(tokens=other.tokens - self.tokens, minus=True) 
            else:
                return TokenSetSymbolic(tokens=other.tokens.union(self.tokens), minus=False) 
    
    def intersect(self, other):
        if other == "∅": return "∅"
        if other == "*": return self

        assert type(other) is TokenSetSymbolic, "Can only intersect two TokenSetSymbolics."

        if self.minusset:
            if other.minusset:
                return TokenSetSymbolic(tokens=self.tokens.union(other.tokens), minus=True)
            else:
               return TokenSetSymbolic(tokens=other.tokens - self.tokens, minus=False) 
        else:
            if other.minusset:
                return TokenSetSymbolic(tokens=self.tokens - other.tokens, minus=False) 
            else:
                return TokenSetSymbolic(tokens=self.tokens.intersection(other.tokens), minus=False)

    def setminus(self, other):
        if other == "*": 
            return "∅"
        if other == "∅" or (not other.minusset and len(other.tokens) == 0):
            return TokenSetSymbolic(tokens=self.tokens, minus=self.minusset)
        
        if self.minusset:
            if other.minusset:
                return TokenSetSymbolic(tokens=self.tokens.union(other.tokens), minus=True)
            else:
                excluded_tokens = self.tokens.union(other.tokens)
                return TokenSetSymbolic(tokens=excluded_tokens, minus=True)
        else:
            if other.minusset:
                return TokenSetSymbolic(tokens=self.tokens.intersection(other.tokens), minus=False) 
            else:
                return TokenSetSymbolic(tokens=self.tokens - other.tokens, minus=False)

    def starts_with(self, s):
        if s in self.tokens: 
            return not self.minusset # returns True if not minusset and s in self.tokens
        else:
            for s in self.tokens:
                if s.startswith(s): 
                    return not self.minusset # returns True if not minusset and some s in self.tokens starts with s
            return self.minusset

    def __len__(self):
        if self.minusset: 
            # cannot determine this without knowledge of the vocabulary size
            return 9999
        else: return len(self.tokens)

    def __repr__(self):
        return str(self)

    def __str__(self):
        tokens_str = "{{{}}}".format(", ".join([t for t in sorted(list(self.tokens))]))

        if self.minusset:
            if len(self.tokens) == 0:
                return "*"
            return "* \ {}".format(tokens_str)
        else:
            if len(self.tokens) == 0: 
                return "{}"
            return tokens_str

    def __eq__(self, other):
        if other == "∅": return False
        if other == "*": return False

        assert type(other) is TokenSetSymbolic, "Can only compare (==) two TokenSetSymbolics."

        return other.minusset == self.minusset and str(sorted(list(self.tokens))) == str(sorted(list(other.tokens)))

TokenSet = TokenSetConcrete

def intersect(*args):
    assert len(args) != 0, "Intersection of zero patterns is not possible."
    if len(args) == 1: return args[0]
    if len(args) != 2: return intersect(args[0], intersect(*args[1:]))
    p1, p2 = args

    if p1 == p2: return p1
    
    if p1 == "∅" or p2 == "∅": return "∅"
    if p1 == "*": return p2
    if p2 == "*": return p1
    
    tokens = p1.intersect(p2)
    
    if len(tokens) == 0: return "∅"

    return tokens

def union(p1, p2):
    if p1 == p2: return p1
    
    if p1 == "*": return p1
    if p2 == "*": return p2
    if p1 == "∅": return p2
    if p2 == "∅": return p1

    return p1.union(p2)

def tset(*tokens, regex=False, prefix=False, exact=False, charlen=None):
    if charlen is not None:
        return TokenSet(charlen=charlen)
    if regex:
        assert len(tokens) == 1, "cannot create a TokenSet from multiple regexes."
        return TokenSet(regex=tokens[0])
    if len(tokens) == 1 and type(tokens[0]) is set:
        return TokenSet(set(list(tokens[0])), minus=False)
    return TokenSet(set(tokens), minus=False, prefix=prefix, exact=exact)

def ntset(*tokens):
    if len(tokens) == 1 and type(tokens[0]) is set:
        return TokenSet(set(list(tokens[0])), minus=True)
    return TokenSet(set(tokens), minus=True)

class ArgTuple(tuple): 
    def __repr__(self) -> str:
        return "ArgTuple" + super().__repr__()

def setminus(p1, p2):
    if p1 == "∅": return "∅"
    if p2 == "∅": return p1
    if p1 == "*": p1 = TokenSet(set([]), minus=True)
    
    assert type(p1) is TokenSet
    r = p1.setminus(p2)

    if type(r) is TokenSet and (type(r) is TokenSetSymbolic and not r.minusset and len(r.tokens) == 0):
        return "∅"
    else:
        return r


# coding=utf-8
# This file is adapted from the original file as part of the huggingface/transformers library, 
# which is licensed and distributed according to the following terms:
#
# Copyright 2020 The Google AI Language Team Authors, Facebook AI Research authors and The HuggingFace Inc. team.
# Copyright (c) 2020, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
from dataclasses import dataclass
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union
import inspect

import torch
import torch.distributed as dist
from torch import nn

from transformers.file_utils import ModelOutput
from transformers.generation.utils import (
    EncoderNoRepeatNGramLogitsProcessor,
    ForcedBOSTokenLogitsProcessor,
    ForcedEOSTokenLogitsProcessor,
    HammingDiversityLogitsProcessor,
    InfNanRemoveLogitsProcessor,
    LogitsProcessorList,
    MinLengthLogitsProcessor,
    NoBadWordsLogitsProcessor,
    NoRepeatNGramLogitsProcessor,
    PrefixConstrainedLogitsProcessor,
    RepetitionPenaltyLogitsProcessor,
    TemperatureLogitsWarper,
    TopKLogitsWarper,
    TopPLogitsWarper,
)
from transformers.generation.stopping_criteria import (
    MaxLengthCriteria,
    MaxNewTokensCriteria,
    MaxTimeCriteria,
    StoppingCriteriaList,
    validate_stopping_criteria,
)
from transformers.utils import logging

from lmql.model.hf_beam_search import BeamSearchScorer, BeamScorer
from lmql.runtime.rewriter import RewrittenInputIds, ActivePromptTokens

logger = logging.get_logger(__name__)

def ensure_tensor(v):
    if torch.is_tensor(v): 
        return v
    else:
        return torch.tensor(v)
class AsyncLogitsProcessorList(list):
    """
    This class can be used to create a list of :class:`~transformers.LogitsProcessor` or
    :class:`~transformers.LogitsWarper` to subsequently process a :obj:`scores` input tensor. This class inherits from
    list and adds a specific `__call__` method to apply each :class:`~transformers.LogitsProcessor` or
    :class:`~transformers.LogitsWarper` to the inputs.
    """

    async def __call__(self, input_ids: torch.LongTensor, scores: torch.FloatTensor, filter_processors=None, **kwargs) -> torch.FloatTensor:
        for processor in self:
            if str(type(processor)) in (filter_processors if filter_processors is not None else []):
                continue
            function_args = inspect.signature(processor.__call__).parameters
            is_async = inspect.iscoroutinefunction(processor) or inspect.iscoroutinefunction(processor.__call__)
            
            if len(function_args) > 2:
                assert all(
                    arg in kwargs for arg in list(function_args.keys())[2:]
                ), f"Make sure that all the required parameters: {list(function_args.keys())} for {processor.__class__} are passed to the logits processor."
                
                if is_async:
                    scores = await processor(input_ids, scores, **kwargs)
                else:
                    scores = processor(input_ids, scores, **kwargs)
            else:
                if is_async:
                    scores = await processor(input_ids, scores)
                else:
                    scores = processor(input_ids, scores)
        return scores

class GenerationMixin:
    """
    A class containing all of the functions supporting generation, to be used as a mixin in
    :class:`~transformers.PreTrainedModel`.
    """

    def right_pad(self, input_ids: List[torch.LongTensor], pad_token_id: int, return_attention_mask: bool = True) -> Tuple[torch.LongTensor, Optional[torch.LongTensor]]:
        lengths = set([len(x) for x in input_ids])
            
        if len(lengths) == 1:
            return {
                "input_ids": torch.stack(input_ids, dim=0)
            }

        attention_mask = torch.stack(
            [torch.tensor([1] * len(x) + [0] * (max(lengths) - len(x)), dtype=torch.long) for x in input_ids]
        )
        input_ids = torch.stack(
            [torch.cat([x, torch.tensor([pad_token_id] * (max(lengths) - len(x)), dtype=torch.long)], axis=0) for x in input_ids]
        )

        return {
            "input_ids": input_ids,
            "attention_mask": attention_mask
        }

    def prepare_inputs_for_generation(self, input_ids: torch.LongTensor, eos_token_id=None, **kwargs) -> Dict[str, Any]:
        """
        Implement in subclasses of :class:`~transformers.PreTrainedModel` for custom behavior to prepare inputs in the
        generate method.
        """
        if torch.is_tensor(input_ids):
            return {
                "input_ids": input_ids
            }
        
        if type(input_ids) is list:
            lengths = set([len(x) for x in input_ids])
            
            if len(lengths) == 1:
                return {
                    "input_ids": torch.stack(input_ids, dim=0)
                }

            attention_mask = torch.stack(
                [torch.tensor([0] * (max(lengths) - len(x)) + [1] * len(x), dtype=torch.long) for x in input_ids]
            )
            input_ids = torch.stack(
                [torch.cat([torch.tensor([eos_token_id] * (max(lengths) - len(x)), dtype=torch.long), x], axis=0) for x in input_ids]
            )

            return {
                "input_ids": input_ids,
                "attention_mask": attention_mask
            }

    def adjust_logits_during_generation(self, logits: torch.FloatTensor, **kwargs) -> torch.FloatTensor:
        """
        Implement in subclasses of :class:`~transformers.PreTrainedModel` for custom behavior to adjust the logits in
        the generate method.
        """
        return logits

    def _prepare_input_ids_for_generation(
        self, bos_token_id: Optional[int], encoder_outputs: Optional[ModelOutput]
    ) -> torch.LongTensor:
        if self.config.is_encoder_decoder and encoder_outputs is not None:
            # make dummy input_ids with value -100, as a sanity check ensuring that they won't be used for encoding
            shape = encoder_outputs.last_hidden_state.size()[:-1]
            return torch.ones(shape, dtype=torch.long, device=self.device) * -100

        if bos_token_id is None:
            raise ValueError("`bos_token_id` has to be defined when no `input_ids` are provided.")
        return torch.ones((1, 1), dtype=torch.long, device=self.device) * bos_token_id

    def _prepare_attention_mask_for_generation(
        self, input_ids: torch.Tensor, pad_token_id: int, eos_token_id: int
    ) -> torch.LongTensor:
        is_pad_token_in_inputs_ids = (pad_token_id is not None) and (pad_token_id in input_ids)
        is_pad_token_not_equal_to_eos_token_id = (eos_token_id is None) or (
            (eos_token_id is not None) and (pad_token_id != eos_token_id)
        )
        if is_pad_token_in_inputs_ids and is_pad_token_not_equal_to_eos_token_id:
            return input_ids.ne(pad_token_id).long()
        return input_ids.new_ones(input_ids.shape, dtype=torch.long)

    def _prepare_encoder_decoder_kwargs_for_generation(
        self, input_ids: torch.LongTensor, model_kwargs
    ) -> Dict[str, Any]:
        if "encoder_outputs" not in model_kwargs:
            # retrieve encoder hidden states
            encoder = self.get_encoder()
            encoder_kwargs = {
                argument: value
                for argument, value in model_kwargs.items()
                if not (argument.startswith("decoder_") or argument.startswith("cross_attn"))
            }
            model_kwargs["encoder_outputs"]: ModelOutput = encoder(input_ids, return_dict=True, **encoder_kwargs)
        return model_kwargs

    def _prepare_decoder_input_ids_for_generation(
        self, input_ids: torch.LongTensor, decoder_start_token_id: int = None, bos_token_id: int = None
    ) -> torch.LongTensor:
        decoder_start_token_id = self._get_decoder_start_token_id(decoder_start_token_id, bos_token_id)
        decoder_input_ids = (
            torch.ones((input_ids.shape[0], 1), dtype=torch.long, device=input_ids.device) * decoder_start_token_id
        )
        return decoder_input_ids

    def _get_pad_token_id(self, pad_token_id: int = None, eos_token_id: int = None) -> int:
        if pad_token_id is None and eos_token_id is not None:
            logger.warning(f"Setting `pad_token_id` to `eos_token_id`:{eos_token_id} for open-end generation.")
            pad_token_id = eos_token_id
        return pad_token_id

    def _get_decoder_start_token_id(self, decoder_start_token_id: int = None, bos_token_id: int = None) -> int:
        decoder_start_token_id = (
            decoder_start_token_id if decoder_start_token_id is not None else self.config.decoder_start_token_id
        )
        bos_token_id = bos_token_id if bos_token_id is not None else self.config.bos_token_id

        if decoder_start_token_id is not None:
            return decoder_start_token_id
        elif (
            hasattr(self.config, "decoder")
            and hasattr(self.config.decoder, "decoder_start_token_id")
            and self.config.decoder.decoder_start_token_id is not None
        ):
            return self.config.decoder.decoder_start_token_id
        elif bos_token_id is not None:
            return bos_token_id
        elif (
            hasattr(self.config, "decoder")
            and hasattr(self.config.decoder, "bos_token_id")
            and self.config.decoder.bos_token_id is not None
        ):
            return self.config.decoder.bos_token_id
        raise ValueError(
            "`decoder_start_token_id` or `bos_token_id` has to be defined for encoder-decoder generation."
        )

    @staticmethod
    def _expand_inputs_for_generation(
        input_ids: torch.LongTensor,
        expand_size: int = 1,
        is_encoder_decoder: bool = False,
        attention_mask: torch.LongTensor = None,
        encoder_outputs: ModelOutput = None,
        **model_kwargs,
    ) -> Tuple[torch.LongTensor, Dict[str, Any]]:
        expanded_return_idx = (
            torch.arange(input_ids.shape[0]).view(-1, 1).repeat(1, expand_size).view(-1).to(input_ids.device)
        )
        input_ids = input_ids.index_select(0, expanded_return_idx)

        if "token_type_ids" in model_kwargs:
            token_type_ids = model_kwargs["token_type_ids"]
            model_kwargs["token_type_ids"] = token_type_ids.index_select(0, expanded_return_idx)

        if attention_mask is not None:
            model_kwargs["attention_mask"] = attention_mask.index_select(0, expanded_return_idx)

        if is_encoder_decoder:
            assert encoder_outputs is not None
            encoder_outputs["last_hidden_state"] = encoder_outputs.last_hidden_state.index_select(
                0, expanded_return_idx.to(encoder_outputs.last_hidden_state.device)
            )
            model_kwargs["encoder_outputs"] = encoder_outputs
        return input_ids, model_kwargs

    @staticmethod
    def _update_model_kwargs_for_generation(
        outputs: ModelOutput, model_kwargs: Dict[str, Any], 
        is_encoder_decoder: bool = False, 
        padding_tokens: Optional[torch.LongTensor] = None
    ) -> Dict[str, Any]:
        # print("TODO: _update_model_kwargs_for_generation still has to be adapted to handle padding_tokens (adjust position_ids, token_type_ids, attention_mask)")

        # update past
        if "past_key_values" in outputs:
            model_kwargs["past"] = outputs.past_key_values
        elif "mems" in outputs:
            model_kwargs["past"] = outputs.mems
        elif "past_buckets_states" in outputs:
            model_kwargs["past"] = outputs.past_buckets_states
        else:
            model_kwargs["past"] = None

        # update token_type_ids with last value
        if "token_type_ids" in model_kwargs:
            token_type_ids = model_kwargs["token_type_ids"]
            model_kwargs["token_type_ids"] = torch.cat([token_type_ids, token_type_ids[:, -1].unsqueeze(-1)], dim=-1)

        # update attention mask
        if not is_encoder_decoder:
            if "attention_mask" in model_kwargs:
                attention_mask = model_kwargs["attention_mask"]
                model_kwargs["attention_mask"] = torch.cat(
                    [attention_mask, attention_mask.new_ones((attention_mask.shape[0], 1))], dim=-1
                )

        return model_kwargs

    def _reorder_cache(self, past, beam_idx):
        raise NotImplementedError(
            f"Make sure that a `_reorder_cache` function is correctly implemented in {self.__class__.__module__} to enable beam search for {self.__class__}"
        )

    def _get_logits_warper(
        self, top_k: int = None, top_p: float = None, temperature: float = None, num_beams: int = None
    ) -> LogitsProcessorList:
        """
        This class returns a :obj:`~transformers.LogitsProcessorList` list object that contains all relevant
        :obj:`~transformers.LogitsWarper` instances used for multinomial sampling.
        """

        # init warp parameters
        top_k = top_k if top_k is not None else self.config.top_k
        top_p = top_p if top_p is not None else self.config.top_p
        temperature = temperature if temperature is not None else self.config.temperature
        # instantiate warpers list
        warpers = LogitsProcessorList()

        # the following idea is largely copied from this PR: https://github.com/huggingface/transformers/pull/5420/files
        # all samplers can be found in `generation_utils_samplers.py`
        if temperature is not None and temperature != 1.0:
            warpers.append(TemperatureLogitsWarper(temperature))
        if top_k is not None and top_k != 0:
            warpers.append(TopKLogitsWarper(top_k=top_k, min_tokens_to_keep=(2 if num_beams > 1 else 1)))
        if top_p is not None and top_p < 1.0:
            warpers.append(TopPLogitsWarper(top_p=top_p, min_tokens_to_keep=(2 if num_beams > 1 else 1)))
        return warpers

    def _get_logits_processor(
        self,
        repetition_penalty: float,
        no_repeat_ngram_size: int,
        encoder_no_repeat_ngram_size: int,
        encoder_input_ids: torch.LongTensor,
        bad_words_ids: List[List[int]],
        min_length: int,
        max_length: int,
        eos_token_id: int,
        forced_bos_token_id: int,
        forced_eos_token_id: int,
        prefix_allowed_tokens_fn: Callable[[int, torch.Tensor], List[int]],
        num_beams: int,
        num_beam_groups: int,
        diversity_penalty: float,
        remove_invalid_values: bool,
        additional_logits_processors: Optional[list],
    ) -> LogitsProcessorList:
        """
        This class returns a :obj:`~transformers.LogitsProcessorList` list object that contains all relevant
        :obj:`~transformers.LogitsProcessor` instances used to modify the scores of the language model head.
        """
        processors = AsyncLogitsProcessorList()

        # init warp parameters
        repetition_penalty = repetition_penalty if repetition_penalty is not None else self.config.repetition_penalty
        no_repeat_ngram_size = (
            no_repeat_ngram_size if no_repeat_ngram_size is not None else self.config.no_repeat_ngram_size
        )
        encoder_no_repeat_ngram_size = (
            encoder_no_repeat_ngram_size
            if encoder_no_repeat_ngram_size is not None
            else self.config.encoder_no_repeat_ngram_size
        )
        bad_words_ids = bad_words_ids if bad_words_ids is not None else self.config.bad_words_ids
        min_length = min_length if min_length is not None else self.config.min_length
        eos_token_id = eos_token_id if eos_token_id is not None else self.config.eos_token_id
        diversity_penalty = diversity_penalty if diversity_penalty is not None else self.config.diversity_penalty
        forced_bos_token_id = (
            forced_bos_token_id if forced_bos_token_id is not None else self.config.forced_bos_token_id
        )
        forced_eos_token_id = (
            forced_eos_token_id if forced_eos_token_id is not None else self.config.forced_eos_token_id
        )
        remove_invalid_values = (
            remove_invalid_values if remove_invalid_values is not None else self.config.remove_invalid_values
        )
        # instantiate processors list

        # the following idea is largely copied from this PR: https://github.com/huggingface/transformers/pull/5420/files
        # all samplers can be found in `generation_utils_samplers.py`
        if diversity_penalty is not None and diversity_penalty > 0.0:
            processors.append(
                HammingDiversityLogitsProcessor(
                    diversity_penalty=diversity_penalty, num_beams=num_beams, num_beam_groups=num_beam_groups
                )
            )
        if repetition_penalty is not None and repetition_penalty != 1.0:
            processors.append(RepetitionPenaltyLogitsProcessor(penalty=repetition_penalty))
        if no_repeat_ngram_size is not None and no_repeat_ngram_size > 0:
            processors.append(NoRepeatNGramLogitsProcessor(no_repeat_ngram_size))
        if encoder_no_repeat_ngram_size is not None and encoder_no_repeat_ngram_size > 0:
            if self.config.is_encoder_decoder:
                processors.append(EncoderNoRepeatNGramLogitsProcessor(encoder_no_repeat_ngram_size, encoder_input_ids))
            else:
                raise ValueError(
                    "It's impossible to use `encoder_no_repeat_ngram_size` with decoder-only architecture"
                )
        if bad_words_ids is not None:
            processors.append(NoBadWordsLogitsProcessor(bad_words_ids, eos_token_id))
        if min_length is not None and eos_token_id is not None and min_length > -1:
            processors.append(MinLengthLogitsProcessor(min_length, eos_token_id))
        if prefix_allowed_tokens_fn is not None:
            processors.append(PrefixConstrainedLogitsProcessor(prefix_allowed_tokens_fn, num_beams // num_beam_groups))
        if forced_bos_token_id is not None:
            processors.append(ForcedBOSTokenLogitsProcessor(forced_bos_token_id))
        if forced_eos_token_id is not None:
            processors.append(ForcedEOSTokenLogitsProcessor(max_length, forced_eos_token_id))
        if remove_invalid_values is True:
            processors.append(InfNanRemoveLogitsProcessor())

        processors += (additional_logits_processors if additional_logits_processors is not None else [])
        return processors

    async def logits(
        self,
        input_ids: torch.LongTensor,
        logits_mask: Optional[torch.LongTensor] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        top_k: Optional[int] = None,
        top_p: Optional[float] = None,
        temperature: Optional[float] = None,
        repetition_penalty=None,
        no_repeat_ngram_size=None,
        bad_words_ids=None,
        prefix_allowed_tokens_fn=None,
        diversity_penalty=None,
        remove_invalid_values=None,
        additional_logits_processors=None,
        min_tokens_to_keep: Optional[int] = None,
        additional_logits_processor_mask: Optional[torch.BoolTensor] = None,
        eos_token_id: Optional[int] = None,
        **model_kwargs,
    ) -> Tuple[torch.FloatTensor, torch.FloatTensor]:
        
        logits_processor = self._get_logits_processor(
            repetition_penalty=repetition_penalty,
            no_repeat_ngram_size=no_repeat_ngram_size,
            encoder_no_repeat_ngram_size=None,
            encoder_input_ids=None,
            bad_words_ids=bad_words_ids,
            min_length=None,
            max_length=None,
            eos_token_id=None,
            forced_bos_token_id=None,
            forced_eos_token_id=None,
            prefix_allowed_tokens_fn=prefix_allowed_tokens_fn,
            num_beams=None,
            num_beam_groups=None,
            diversity_penalty=diversity_penalty,
            remove_invalid_values=remove_invalid_values,
            additional_logits_processors=additional_logits_processors
        )

        assert min_tokens_to_keep is None or min_tokens_to_keep == 1 or min_tokens_to_keep == 2, "min_tokens_to_keep must be either 1,2 or None"
        # get probability distribution warper
        logits_warper = self._get_logits_warper(
            top_k=top_k, top_p=top_p, temperature=temperature, num_beams=2 if min_tokens_to_keep == 2 else 1
        )
        
        output_attentions = output_attentions if output_attentions is not None else self.config.output_attentions
        output_hidden_states = (
            output_hidden_states if output_hidden_states is not None else self.config.output_hidden_states
        )

        # prepare model inputs
        model_inputs = self.prepare_inputs_for_generation(input_ids, **model_kwargs, eos_token_id=eos_token_id)

        # forward pass to get next token
        outputs = await self(
            **model_inputs,
            return_dict=True,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
        )
        next_token_logits = outputs.logits[:, -1, :]
        raw_logits = next_token_logits.clone().detach()
        
        next_token_scores = await logits_processor(input_ids, next_token_logits, additional_logits_processor_mask=additional_logits_processor_mask)
        
        # apply additive logits mask
        if logits_mask is not None:
            assert additional_logits_processors is None, "additional_logits_processors and the additive logits_mask are expected to be mutually exclusive"
            tokenizer_range = logits_mask.size(-1)
            next_token_scores[:,:tokenizer_range] = next_token_scores[:,:tokenizer_range] + logits_mask
        
        next_token_scores = logits_warper(input_ids, next_token_scores)

        next_token_scores = next_token_scores - next_token_scores.logsumexp(-1, keepdim=True)
        raw_next_token_scores = raw_logits - raw_logits.logsumexp(-1, keepdim=True)

        return next_token_scores, raw_next_token_scores
    

    async def score(
        self,
        input_ids: torch.LongTensor,
        completion: torch.LongTensor,
        eos_token_id: Optional[int] = None,
        **model_kwargs,
    ) -> Tuple[torch.FloatTensor, torch.FloatTensor]:
        # prepare model inputs
        model_inputs = self.prepare_inputs_for_generation(input_ids, **model_kwargs, eos_token_id=eos_token_id)

        num_tokens_to_score = completion.size(-1)
        logits = []
        token_scores = []
        
        for i in range(num_tokens_to_score):
            # forward pass to get next token
            outputs = await self(
                **model_inputs,
                return_dict=True,
                output_attentions=False,
                output_hidden_states=False,
            )

            next_token_logits = outputs.logits[:, -1, :]
            logits.append(next_token_logits)
            next_token_logits = torch.log_softmax(next_token_logits, dim=-1)
            token_scores.append(next_token_logits.gather(-1, completion[:, i].unsqueeze(-1)))

            model_inputs["input_ids"] = torch.cat([model_inputs["input_ids"], completion[:, i].unsqueeze(-1)], dim=-1)

        return torch.cat(token_scores, dim=-1), torch.cat([l.unsqueeze(1) for l in logits], dim=1)

    async def unbiased_next_token_beam_scores(self, next_token_scores, beam_scores):
        # active_prompting_token_score = await logits_processor(input_ids, next_token_scores, filter_processors=["<class 'function'>"])
        unbiased_next_token_score = next_token_scores.clone()
        unbiased_next_token_score = unbiased_next_token_score + beam_scores[:, None].expand_as(unbiased_next_token_score)
        # unbiased_next_token_score = logits_warper(input_ids, unbiased_next_token_score)
        # unbiased_next_token_score = torch.zeros_like(next_token_scores) + beam_scores[:, None].expand_as(next_token_scores)

        return unbiased_next_token_score



import asyncio
import os
import inspect
import lmql.runtime.bopenai as openai
from lmql.runtime.stats import Stats
from dataclasses import dataclass
from typing import Any, Callable, Optional, List, Union

from collections import namedtuple
import numpy as np
from lmql.utils import nputil

from lmql.runtime.decoder_head import DecoderHead
from lmql.runtime.rewriter import InputIdRewriter, ActivePromptingRewriter

from lmql.runtime.tokenizer import load_tokenizer
from lmql.runtime.dclib.lmql_adapter import QueryDcLibAdapter
from lmql.runtime.dclib.dclib_model import DcModel
from lmql.runtime.dclib.dclib_seq import deepcopy, deepmerge, is_deterministic, detseq, DecoderSequence
import lmql.runtime.dclib as dc

def is_allowed(m): 
    """
    Given a logits mask, sets tensor cell value to True iff the corresponding token is allowed according to the mask.
    """
    return np.isclose(m, 0, atol=1e-8)

def openai_complete_create(*args, **kwargs):
    if kwargs.get("chaos", False):
        import random
        if random.random() < 0.5:
            if random.random() < 0.5:
                raise openai.error.ServiceUnavailableError("Chaos monkey error", 500, "Chaos monkey error")
            else:
                raise openai.error.APIError("Chaos monkey error", 500, "Chaos monkey error")

    return openai.Completion.create(*args, **kwargs)

@dataclass
class CompleteTask:
    op: Callable
    result: Any
    continuation_type: str
    logit_mask_or_fixed_id: Optional[Union[np.ndarray, int]] = None

    async def run(self, retries):
        try:
            return CompleteTask(self.op, await self.op(), self.continuation_type, self.logit_mask_or_fixed_id)
        except Exception as e:
            if type(e) is not openai.error.APIError and type(e) is not openai.error.ServiceUnavailableError:
                raise e
            if retries > 0:
                print("OpenAI API error: ", e, ". Retrying...", flush=True)
                print("Retrying operation due to OpenAI API error", e)
                return await self.run(retries - 1)
            else:
                print("Exceeded ", retries, " retries. Giving up.", flush=True)
                raise e

@dataclass
class CompletionResult:
    buffer: openai.response_buffer
    continuation_type: str
    logit_mask_or_fixed_id: Optional[Union[np.ndarray, int]] = None

@dataclass
class CompletionCall:
    mode: str
    logit_mask_or_fixed_id: np.ndarray
    input_ids: np.ndarray
    kwargs: Any
    stopping_phrases: List[str] = None
    
    # true iff inverting the api_mask leads to a smaller mask
    invert: bool = False
    
    @property
    def api_mask(self, invert=None):
        mask = self.logit_mask_or_fixed_id
        assert nputil.is_array(mask), "api_mask(): logit_mask_or_fixed_id must be a LongTensor not a " + str(type(mask))
        
        if self.invert:
            masked = (mask >= 0)
        else:
            masked = (mask < 0)
        mask_value = 100 if self.invert else -100
        return {int(idx): mask_value for idx in np.nonzero(masked)[0]}

    @property
    def continuation_type(self):
        if self.mode == "fixed":
            return None
        # unique cache key identifying the type of this completion
        parameter_values_key_segment = "temp-" + str(self.kwargs["temperature"]) + "-logprobs-" + str(self.kwargs["logprobs"])
        if self.mode == "complete":
            api_mask = self.api_mask
            mask_key_segment = [f"{id}={value}" for id, value in sorted(api_mask.items(), key=lambda x: x[0])]
            mask_key_segment = "-".join(mask_key_segment)
        else:
            mask_key_segment = "*"
        return f"{parameter_values_key_segment}-{mask_key_segment}"

class DclibOpenAiModel(DcModel):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, truncation_threshold=-120, init_workers=False, **kwargs)

        # if available, store reference to output writer for eager stats reporting
        self.output_writer = None
        if "output_writer" in kwargs:
            self.output_writer = kwargs["output_writer"]
        
        self.model_identifier = "openai/" + self.model.model_identifier

        self.model.chunk_size = kwargs.get("openai_chunksize", 32)
        self.num_billed_tokens = {}
        self.num_requests = 0

        self.stats = Stats("openai")

        # openai.Completion.set_chaos(0.05)
        openai.AsyncConfiguration.set_tokenizer(self.tokenize)

    def log_billable_tokens(self, n: int):
        pass # openai keeps track of billable tokens vai bopenai
    
    def log_queries(self, n: int):
        pass # openai keeps track of queries via bopenai

    def prepare_completion_call(self, s, mask, **kwargs):
        """
        Computes an API compatible mask from the provided logit mask, as well as the required mode of completion.

        Returns Values:
            - (mask_dict, "*"): Complete with the full vocabulary (no logit_bias necessary)
            - (mask_dict, "complete"): Complete with the provided mask mask_dict, to be passed like this to the OpenAI endpoint.
            - (token_id, "fixed"): Complete with the provided token_id with full probability 1.0 (no API use necessary)

        """
        stopping_phrases = s.data("head.stopping_phrases.text")

        if mask is None:
            return CompletionCall("*", None, s.input_ids, kwargs, stopping_phrases=stopping_phrases)

        invert = False
        num_allowed = is_allowed(mask).sum(axis=-1)
        assert num_allowed > 0, "DclibOpenAiModel: encountered logits mask with no allowed tokens"

        if num_allowed == 1:
            # check for <eos> case
            if is_allowed(mask[self.eos_token_id]):
                return CompletionCall("fixed", self.eos_token_id, s.input_ids, kwargs, stopping_phrases=stopping_phrases)
            else:
                # otherwise we can treat this as a score call
                return CompletionCall("fixed", mask.argmax(axis=-1), s.input_ids, kwargs, stopping_phrases=stopping_phrases)
        elif num_allowed < mask.shape[-1]:
            if mask.shape[-1] - num_allowed > num_allowed:
                # if we have to mask more than half of the tokens, we should just invert the masking
                invert = True
        else: # num_allowed == mask.shape[-1] (full vocabulary)
            return CompletionCall("*", None, s.input_ids, kwargs, stopping_phrases=stopping_phrases)

        # num_allowed < mask.shape[-1] and num_allowed > 1 (needs mask)
        return CompletionCall("complete", mask, s.input_ids, kwargs, invert=invert, stopping_phrases=stopping_phrases)

    async def api_score(self, input_ids, offset):
        kwargs = {
            "model": self.model.model_identifier,
            "prompt": input_ids.tolist(),
            "max_tokens": 0,
            "temperature": 0,
            "logprobs": 1,
            "user": "lmql",
            "echo": True
        }

        logprobs = []
        async for data in await openai.Completion.create(**kwargs):
            logprobs += data["logprobs"]["token_logprobs"]
        return np.array(logprobs[offset:], dtype=np.float32)

    async def queue_api_score(self, kwargs):
        # put task in self.score_queue and await result
        loop = asyncio.get_running_loop()
        result_fut = loop.create_future()
        self.score_queue.put_nowait((kwargs,result_fut))
        return await result_fut

    async def _score_next_tokens(self, s, next_tokens):
        return (await self.api_score(np.concatenate([s.input_ids, next_tokens], axis=0), len(s.input_ids)))
    
    async def score(self, sqs: List[DecoderSequence], tokens: List[List[int]], max_batch_size=4, deterministic=False, stop_phrase=False, needs_rewrite=True):
        assert len(sqs) == len(tokens), "Number of sequences and number of tokens to be scored must match, but got {} and {}".format(len(sqs), len(tokens))
        
        completion = [np.array(cont) for cont in tokens]

        def make_detseq(s, token_score, completion):
            return detseq(ids=np.concatenate([s.input_ids, completion[:1]], axis=0), 
                    next_ids=completion[1:],
                    logprobs=np.concatenate([s.logprobs, token_score[:1]], axis=0),
                    next_logprobs=token_score[1:],
                    deterministic=np.concatenate([s.deterministic, np.array([deterministic])]),
                    next_deterministic=np.array([deterministic] * len(completion[1:])),
                    predecessor=s,
                    user_data=None,
                    stop_phrase=np.concatenate([s.stop_phrase, np.array([stop_phrase])]),
                    needs_rewrite=needs_rewrite,
                    sticky_user_data_keys=s.sticky_user_data_keys)
        results = []

        for s,compl,result in zip(sqs, completion, await asyncio.gather(*(self._score_next_tokens(s, compl) for s, compl in zip(sqs, completion)))):
            results.append(make_detseq(s, result, compl))

        return results

    async def async_complete(self, completion_call: Union[CompletionCall, List[CompletionCall]], **kwargs) -> openai.response_buffer:
        assert type(completion_call) is CompletionCall

        batch_size = 1
        input_ids = completion_call.input_ids.reshape(-1)
        assert input_ids.ndim == 1, f"_complete expects input_ids to be a 1D tensor when only one completion_call is passed, got {input_ids.ndim}D tensor."

        temperature = completion_call.kwargs.get("temperature", 0.0)
        logprobs = completion_call.kwargs.get("logprobs", 5)

        kwargs = {
            "model": self.model.model_identifier,
            "prompt": input_ids.tolist(), # no more batching at this point
            "max_tokens": self.model.chunk_size,
            "temperature": temperature,
            "logprobs": logprobs,
            "user": "lmql",
            "stream": True,
            "echo": True
        }

        mode = completion_call.mode
        
        if mode == "*": # complete without mask
            pass
        elif mode == "complete": # complete with mask
            logit_bias = completion_call.api_mask
            kwargs.update({"logit_bias": logit_bias})
        elif mode == "fixed": # complete with fixed token
            fixed_next_token = completion_call.logit_mask_or_fixed_id # special return value case for prepare function
            # TODO revisit this, what kind of probability do we want here (masked or unmasked/scored)
            if fixed_next_token == self.eos_token_id:
                return CompletionResult(openai.response_buffer.singleton(token=fixed_next_token, token_logprob=0), completion_call.continuation_type, completion_call.logit_mask_or_fixed_id)
            else:
                fixed_next_token = nputil.ensure_array(fixed_next_token, dtype=np.int64)
                logprob = (await self.api_score(np.concatenate([input_ids, fixed_next_token.reshape(1)], axis=0), len(input_ids)))
                return CompletionResult(openai.response_buffer.singleton(token=fixed_next_token, token_logprob=logprob), completion_call.continuation_type, completion_call.logit_mask_or_fixed_id)
        elif len(completion_call.stopping_phrases) > 0:
            if len(completion_call.stopping_phrases) > 4:
                # same but blaming it more on OpenAI
                print("warning: the number of stopping phrases that would need to be passed to the OpenAI API is greater than 4. Since the OpenAI API only supports up to 4 stopping phrases, the first 4 stopping phrases will be passed to the API. Other stopping phrases will also be enforced, but may lead to an increase in the number of tokens billed to the user.")
            kwargs.update({"stop": completion_call.stopping_phrases[:4]})
        else:
            assert False, f"Internal openai API dispatcher returned an unknown completion mode {mode}"

        # TODO: we are now overestimate the number of tokens billed to the user since we are not account for stopping phrases for the sake of streaming
        self.count_billed_tokens(input_ids.size + kwargs.get("max_tokens") * batch_size, self.model_identifier)
        
        if self.model_args.get("chatty_openai", False):
            print("Completion with", kwargs)

        return CompletionResult((await openai.async_buffer(await openai.Completion.create(**kwargs), tokenizer=self.tokenize_list))[input_ids.size:], completion_call.continuation_type, completion_call.logit_mask_or_fixed_id)
    
    async def tokenize_list(self, tokens: List[str]):
        if len(tokens) > 0 and type(tokens[0]) is str:
            return [[t[0]] for t in await self.model.tokenize(tokens)]
        return tokens
    
    async def _complete(self, completion_call: Union[CompletionCall, List[CompletionCall]], **kwargs):
        if type(completion_call) is list:
            input_ids = np.stack([c.input_ids for c in completion_call], axis=0)
            assert input_ids.ndim == 2, f"_complete expects input_ids to be a 1D tensor per completion call, when multiple completion calls are passed, got {input_ids.ndim}D tensor."
            # all other call parameters are assumed to be the same
            batch_size = len(completion_call)
            completion_call = completion_call[0]
        else:
            batch_size = 1
            input_ids = completion_call.input_ids
            assert input_ids.ndim == 1, f"_complete expects input_ids to be a 1D tensor when only one completion_call is passed, got {input_ids.ndim}D tensor."

        temperature = completion_call.kwargs.get("temperature", 0.0)
        logprobs = completion_call.kwargs.get("logprobs", 5)

        kwargs = {
            "model": self.model.model_identifier,
            "prompt": input_ids.tolist()[0], # no more batching at this point
            "max_tokens": self.model.chunk_size,
            "temperature": temperature,
            "logprobs": logprobs,
            "user": "lmql",
            "stream": True,
        }

        mode = completion_call.mode
        
        if mode == "*": # complete without mask
            pass
        elif mode == "complete": # complete with mask
            logit_bias = completion_call.api_mask
            kwargs.update({"logit_bias": logit_bias})
        elif mode == "fixed": # complete with fixed token
            fixed_next_token = completion_call.logit_mask_or_fixed_id # special return value case for prepare function
            # TODO revisit this, what kind of probability do we want here (masked or unmasked/scored)
            if fixed_next_token == self.eos_token_id:
                return OpenAIModelOutputBuffer.fixed_output(fixed_next_token, 0)
            else:
                if not nputil.is_array(fixed_next_token): 
                    fixed_next_token = np.array(fixed_next_token)
                logprob = (await self.api_score(np.concatenate([input_ids, fixed_next_token.reshape(1)], axis=0), len(input_ids))).result
                return OpenAIModelOutputBuffer.fixed_output(fixed_next_token, logprob)
        elif len(completion_call.stopping_phrases) > 0:
            if len(completion_call.stopping_phrases) > 4:
                # same but blaming it more on OpenAI
                print("warning: the number of stopping phrases that would need to be passed to the OpenAI API is greater than 4. Since the OpenAI API only supports up to 4 stopping phrases, the first 4 stopping phrases will be passed to the API. Other stopping phrases will also be enforced, but may lead to an increase in the number of tokens billed to the user.")
            kwargs.update({"stop": completion_call.stopping_phrases[:4]})
        else:
            assert False, f"Internal openai API dispatcher returned an unknown completion mode {mode}"

        # TODO: we are now overestimate the number of tokens billed to the user since we are not account for stopping phrases for the sake of streaming
        self.count_billed_tokens(input_ids.size + kwargs.get("max_tokens") * batch_size, self.model_identifier)
        
        if self.model_args.get("chatty_openai", False):
            print("Completion with", kwargs)
        
        async def complete_op():
            return openai_complete_create(**kwargs)
        return await CompleteTask(complete_op, None, continuation_type=completion_call.continuation_type, 
            logit_mask_or_fixed_id=completion_call.logit_mask_or_fixed_id).run(retries=3)

    def count_billed_tokens(self, n, model):
        if model not in self.num_billed_tokens.keys():
            self.num_billed_tokens[model] = 0
        self.num_billed_tokens[model] += n
        self.num_requests += 1

    async def completion_buffer(self, seqs, temperature=1, **kwargs):
        kwargs.update({"temperature": temperature})
        
        async def get_buffer(i, s):
            with self.stats.timer("logit_masks"):
                # print("completion_buffer", s)
                constrained_seqs = np.array([s.is_query_constrained], dtype=np.bool_)
                logits_mask_result = await self.compute_logits_mask(s.input_ids.reshape(1, -1), [s.user_data], constrained_seqs, **kwargs)
                logits_mask = logits_mask_result.logits_mask[0]

            # update user data with new information obtained when computing logits masks
            if s.user_data is None:
                s.user_data = {}
            s.user_data = deepmerge(deepcopy(s.user_data), logits_mask_result.user_data[0])
            s.user_data["set_by"] = "where"

            completion_call = self.prepare_completion_call(s, logits_mask, **kwargs)

            # if no masking is required, we can use cached continuations if available
            if s.data("openai-continuations") is not None:
                continuations: CompletionResult = s.data("openai-continuations")
                continuation_type = completion_call.continuation_type

                if continuation_type in continuations:
                    continuation = continuations[continuation_type]
                    if await continuation.buffer.empty():
                        # remove continuation (it's empty)
                        del s.data("openai-continuations")[continuation_type]
                    else:
                        # use continuation instead of issuing a new API call
                        return continuation

            # handle deterministic sequences
            if is_deterministic(s) and len(s.next_ids) > 0:
                return CompletionResult(
                    openai.response_buffer.singleton(token=s.next_ids[0], token_logprob=s.next_logprobs[0]),
                    None,
                    None,
                )

            return await self.async_complete(completion_call)

        return await asyncio.gather(*[get_buffer(i, s) for i, s in enumerate(seqs)])

    async def argmax(self, sequences, **kwargs):
        return await self.sample(sequences, num_samples=1, temperature=0)

    def report_stats(self, printer, decoder_step=None):
        if printer is None:
            print("no printer")
            return
        if hasattr(printer, "report_model_stats"):
            s = openai.Completion.get_stats()
            data = {
                "tokens": s.tokens,
                "model": self.model_identifier,
                "req.": s.requests,
                "avb": f"{float(s.sum_batch_size)/max(1,s.requests):.2f}",
            }
            if decoder_step is not None:
                data["_step"] = decoder_step
            printer.report_model_stats(**data)

    async def sample(self, sequences, num_samples=1, **kwargs):
        """
        Returns a pool with `n` sampled successor nodes per node in the pool.
        """
        kwargs = {**self.model_args, **kwargs}

        async def op_sample(seqs):
            completions: List[CompletionResult] = await self.completion_buffer(seqs, logprobs=5, **kwargs)
            
            next_token_ids = []
            next_token_scores = []
            logits = []
            continuation_buffers: List[CompletionResult] = []

            for s, completion in zip(seqs, completions):
                assert not await completion.buffer.empty(), "Completion buffer is empty {}".format(completion.buffer)
                complete_data = (await completion.buffer.get(0))

                # store pointer to continuation buffers in newly expanded nodes
                continuation = CompletionResult(completion.buffer[1:], completion.continuation_type, completion.logit_mask_or_fixed_id)
                continuation_buffers.append(continuation)

                # detect fixed results (i.e. deterministic tokens)
                if "fixed" in complete_data.keys():
                    next_token = complete_data["logprobs"]["tokens"]
                    next_token_score = complete_data["logprobs"]["token_logprobs"]
                    next_token_ids.append(np.array([next_token], dtype=np.int64))
                    next_token_scores.append(np.array([next_token_score], dtype=np.float32))
                    
                    full_logits = np.ones(self.model.tokenizer.vocab_size) * np.finfo(np.float32).min
                    if next_token < len(full_logits):
                        full_logits[next_token] = next_token_score
                    # else: 
                    #  next_token is a special token, which is not in the vocab
                    logits.append(full_logits)
                    continue

                # get sampled token and score
                next_token = complete_data["logprobs"]["tokens"][0]
                next_token_score = complete_data["logprobs"]["token_logprobs"]
                
                probs = sorted(list(complete_data["logprobs"]["top_logprobs"].items()))
                logprobs = [p[1] for p in probs]
                tokens = [p[0] for p in probs]
                prob_tokens = await self.tokenize_list(tokens)
                token_ids = np.array(prob_tokens, dtype=np.int64).reshape(-1)

                full_logits = np.ones(self.model.tokenizer.vocab_size) * np.finfo(np.float32).min
                full_logits[token_ids] = np.array(logprobs)
                full_logits[next_token] = np.finfo(np.float32).min
                assert kwargs.get("temperature", 1.0) != 0.0 or np.all(full_logits < next_token_score), "next token score is not the highest"

                # retroactively apply logits mask to logits
                mask = completion.logit_mask_or_fixed_id
                if mask is None: 
                    pass
                elif type(mask) is int: 
                    full_logits[mask] = np.finfo(np.float32).min
                else: 
                    full_logits[mask < 0] = np.finfo(np.float32).min

                additional_sampled_token_ids, _ = nputil.multinomial(full_logits, num_samples=num_samples - 1)

                seq_next_token_ids = np.concatenate([np.array(next_token).reshape(1), additional_sampled_token_ids], axis=0)
                full_logits[next_token] = next_token_score
                # seq_next_token_scores = full_logits.gather(-1, seq_next_token_ids)
                seq_next_token_scores = np.take_along_axis(full_logits, seq_next_token_ids, axis=-1)

                next_token_ids.append(seq_next_token_ids)
                next_token_scores.append(seq_next_token_scores)
                logits.append(full_logits)

            logits = logits
            next_token_ids = next_token_ids
            next_token_scores = next_token_scores

            def successor_user_data(continuation_buffer: SequenceResult, num_successors):
                default_user_data = {"head": {"variable": s.data("head.variable"), "head_index": s.data("head.head_index")}}
                if continuation_buffer.continuation_type is None:
                    return [default_user_data.copy()] * num_successors
                continuation_as_user_data = {
                    "openai-continuations": {
                        continuation_buffer.continuation_type: continuation_buffer
                    },
                    **(default_user_data.copy())
                }
                return [continuation_as_user_data] + [default_user_data.copy()] * (num_successors - 1)

            s = [s.make_successors(next_token_ids[i], next_token_scores[i], logits=logits[i], 
                user_data=successor_user_data(continuation_buffers[i], len(next_token_ids[i]))) for i,s in enumerate(seqs)]
            return s
        with self.stats.timer("sample"):
            return await sequences.aelement_wise(op_sample)

    async def topk_continuations(self, sequences, k, **kwargs):
        """
        Returns a pool with `n` sampled successor nodes per node in the pool.
        """
        assert k <= 5, "The OpenAI API only supports topk probabilities with k <= 5"
        assert k >= 1, "topk_continuations() requires k >= 1"
        
        assert not "turbo" in self.model_identifier, f"Chat API models do not support topk_continuations which is required for the requested decoding algorithm, use 'sample' or 'argmax' instead."

        kwargs = {**self.model_args, **kwargs}
        kwargs.update({"temperature": 0.0})
        
        async def op_topk(seqs):
            completions: List[CompletionResult] = await self.completion_buffer(seqs, logprobs=k, **kwargs)
            
            next_token_ids = []
            next_token_scores = []
            logits = []
            continuation_buffers: List[CompletionResult] = []

            for s, completion in zip(seqs, completions):
                complete_data = (await completion.buffer.get(0))
                # store pointer to continuation buffers in newly expanded nodes
                continuation = CompletionResult(completion.buffer[1:], completion.continuation_type, completion.logit_mask_or_fixed_id)
                continuation_buffers.append(continuation)

                # detect fixed results (i.e. deterministic tokens)
                if "fixed" in complete_data.keys():
                    next_token = complete_data["logprobs"]["tokens"]
                    next_token_score = complete_data["logprobs"]["token_logprobs"]
                    next_token_ids.append(np.array([next_token], dtype=np.int64))
                    next_token_scores.append(np.array([next_token_score], dtype=np.float32))
                    
                    full_logits = np.ones(self.model.tokenizer.vocab_size) * np.finfo(np.float32).min
                    full_logits[next_token] = next_token_score
                    logits.append(full_logits)
                    continue

                # get sampled token and score
                next_token = complete_data["logprobs"]["tokens"][0]
                next_token_score = complete_data["logprobs"]["token_logprobs"]
                
                probs = sorted(list(complete_data["logprobs"]["top_logprobs"].items()), key=lambda x: x[1], reverse=True)
                logprobs = [p[1] for p in probs]
                tokens = [p[0] for p in probs]
                token_ids = np.array([self.model.tokenizer(t)["input_ids"][0] for t in tokens], dtype=np.int64)

                assert token_ids[0] == next_token, f"top1 logprob token is not the same as the predicted token {token_ids[0]} (top1) != {next_token} (predicted)"

                full_logits = np.ones(self.model.tokenizer.vocab_size) * np.finfo(np.float32).min
                full_logits[token_ids] = np.array(logprobs)

                # retroactively apply logits mask to logits
                mask = completion.logit_mask_or_fixed_id
                if mask is None: pass
                elif type(mask) is int: full_logits[mask] = np.finfo(np.float32).min
                else: full_logits[mask < 0] = np.finfo(np.float32).min
                
                # re-determine logprobs with logits mask applied
                logprobs = np.take_along_axis(full_logits, token_ids, axis=-1)

                # make sure all token_ids are different
                assert len(token_ids) == len(set(token_ids)), "token_ids are not unique"

                next_token_ids.append(token_ids)
                next_token_scores.append(logprobs)
                logits.append(full_logits)

            def successor_user_data(continuation_buffer: SequenceResult, num_successors):
                default_user_data = {"head": {"variable": s.data("head.variable"), "head_index": s.data("head.head_index")}}
                if continuation_buffer.continuation_type is None:
                    return [default_user_data.copy()] * num_successors
                continuation_as_user_data = {
                    "openai-continuations": {
                        continuation_buffer.continuation_type: continuation_buffer
                    },
                    **(default_user_data.copy())
                }
                return [continuation_as_user_data] + [default_user_data.copy()] * (num_successors - 1)

            return [s.make_successors(next_token_ids[i], next_token_scores[i], logits=logits[i], 
                user_data=successor_user_data(continuation_buffers[i], len(logits[i]))) for i,s in enumerate(seqs)]
        
        with self.stats.timer("topk"):
            return await sequences.aelement_wise(op_topk)

class Struct:
    def __init__(self, **entries):
        for k, v in entries.items():
            if isinstance(v, dict):
                entries[k] = Struct(**v)
            if isinstance(v, list):
                entries[k] = [Struct(**i) if isinstance(i, dict) else i for i in v]
        self.__dict__.update(entries)
    
    def __getitem__(self, k):
        return self.__dict__[k]

    # make json serializable
    def __repr__(self):
        return str(self.__dict__)

class OpenAIModelOutputBuffer:
    def __init__(self, complete_result: CompleteTask, n, tokenizer):
        self.n = n
        self.complete_result: CompleteTask = complete_result
        self.res = complete_result.result
        self.tokenizer = tokenizer
        
        self.tokens = [[] for _ in range(n)]
        self.logprobs = [[] for _ in range(n)]
        self.response_data = [[] for _ in range(n)]

        self.preview = None

    @staticmethod
    def fixed_output(token_id, logprob):
        async def data():
            yield Struct(**{
                "tokens": [token_id],
                "logprobs": {
                    "tokens": [token_id],
                    "token_logprobs": [logprob]
                },
                "fixed": True
            })

        return CompleteTask(lambda: data(), data(), continuation_type=None, logit_mask_or_fixed_id=token_id)

    def text_to_tokens(self, l):
        if len(l) == 0 and type(l) is list:
            return []
        if type(l[0]) is int:
            # handles fixed_output case
            return l
        else:
            # handles actual API case
            text = "".join(l)
            return self.tokenizer(text)["input_ids"]

    def skip_items(self, skip=None):
        if skip is None:
            return

        def minimum_token_count_reached():
            return all(len(t) >= skip[i] for i,t in enumerate(self.tokens))

        # skip as many items as needed to reach the previous token count per sequence
        while not minimum_token_count_reached():
            self.process_next_result(next(self.res))

    def process_next_result(self, c):
        # print("data is", data)
        # choices = data.choices
        # assert len(choices) == self.n, f"OpenAI model returned different number of choices than expected: {len(choices)} != {self.n}\n\n {json.dumps(data)}"
        # index = int(c["index"])
        index = 0
        assert self.n == 1, "ModelOutputBuffer only supports n=1"
        self.tokens[index] += self.text_to_tokens(c.logprobs.tokens)
        self.logprobs[index] += c.logprobs.token_logprobs
        self.response_data[index] += [c] * len(c.logprobs.token_logprobs)

    async def advance_stream_iter(self, skip=None):
        try:
            # check if self.res is still coroutine
            if inspect.iscoroutine(self.res):
                self.res = aiter(await self.res)
            # self.skip_items(skip)
            self.process_next_result(await anext(self.res))
            return False
        except StopIteration:
            return True
        except openai.error.APIError as e:
            print("OpenAI API error: ", e, ". Retrying...", flush=True)
            # print("retry with", self.complete_result, "and offset", len(self.tokens), flush=True)
            # re-run underlying complete operation
            self.complete_result = await self.complete_result.run(retries=1)
            self.res = self.complete_result.result
            
            return await self.advance_stream_iter(skip=[len(t) for t in self.tokens])
        except openai.error.ServiceUnavailableError:
            print("OpenAI API Overloaded error: ", e, ". Retrying...", flush=True)
            # print("retry with", self.complete_result, "and offset", len(self.tokens), flush=True)
            # re-run underlying complete operation
            self.complete_result = await self.complete_result.run(retries=1)
            self.res = self.complete_result.result
            
            return await self.advance_stream_iter(skip=[len(t) for t in self.tokens])

    def buffer(self, seq_idx):
        return SequenceResultBuffer(seq_idx, self)

    def view(self, seq_idx):
        return SequenceResult(self, seq_idx)

class FixedBuffer:
    def __init__(self, values):
        self.values = values
    
    async def pop(self):
        return self.values.pop(0)
    
    async def at_end(self):
        return len(self.values) == 0

def fixed(values):
    return FixedBuffer(values)

class EmptyBuffer:
    def at_end(self):
        return True

    async def pop(self):
        assert False, "cannot pop() on empty buffer"

def empty():
    return EmptyBuffer()

class SequenceResultBuffer:
    def __init__(self, i: int, request_buffer: OpenAIModelOutputBuffer):
        self.request_buffer = request_buffer
        self.i = i
        self.items = []
    
    async def pop(self):
        await self.advance_stream_iter(self.id)
        if len(self.request_buffer.tokens[self.i]) == 0:
            raise StopIteration(f"No more tokens in buffer for sequence index {self.i}.")
        return self.request_buffer.tokens[self.i].pop(0), self.request_buffer.logprobs[self.i].pop(0)

    async def advance_stream_iter(self, index):
        # advance stream iter until we have at least one token for this sequence 
        # or the stream ends
        is_done = False
        while not is_done and len(self.request_buffer.tokens[index]) == 0:
            is_done = await self.request_buffer.advance_stream_iter()

    async def at_end(self):
        await self.advance_stream_iter(self.i)
        return len(self.request_buffer.tokens[self.i]) == 0

@dataclass
class OpenAITokenResult:
    token: int
    logprobs: Any
    data: Any

class SequenceResult:
    def __init__(self, request_buffer, result_index: int, offset: int = 0, length: int = None):
        self.request_buffer: OpenAIModelOutputBuffer = request_buffer
        self.result_index = result_index
        self.offset = offset
        self.length = length

        self.collected_data = []
        self.collected_tokens = []
        self.collected_logprobs = []

    @property
    def continuation_type(self):
        return self.request_buffer.complete_result.continuation_type

    def __getitem__(self, key):
        if type(key) is slice:
            return self.slice(key.start, key.stop)
        elif type(key) is int:
            return self.get(key)
        else:
            raise TypeError(f"Can only subscript SequenceResult with int or slice, not {type(key)}")
    
    async def json(self):
        return f"<SequenceResult current_len={self.current_len}>"

    @property
    def current_len(self):
        return max(0, len(self.request_buffer.tokens[self.result_index]) - self.offset)
    

    async def get(self, i: int):
        i = self.offset + i
        while len(self.request_buffer.response_data[self.result_index]) <= i:
            await self.request_buffer.advance_stream_iter()
        
        return OpenAITokenResult(self.request_buffer.tokens[self.result_index][i],
                                 self.request_buffer.logprobs[self.result_index][i],
                                 self.request_buffer.response_data[self.result_index][i])

    def slice(self, lower, upper):
        assert upper is None, "cannot specify upper bound for SequenceResult slice"
        return SequenceResult(self.request_buffer, self.result_index, self.offset + lower, upper)

    async def empty(self):
        if len(self.request_buffer.tokens[self.result_index]) > self.offset:
            return False
        # iterate stream until current_len is greater 0 or the stream ends
        while self.current_len == 0:
            at_end = await self.request_buffer.advance_stream_iter()
            if at_end: break
        return self.current_len == 0

class OptimisticChunkBasedOpenAIModel:
    def __init__(self, model_identifier, tokenizer):
        self.model_identifier = model_identifier.split("openai/",1)[1]
        self.chunk_size = 32
        self.tokenizer = tokenizer

        self.input_ids = []
        self.next_token_scores = []
        self.logits_mask = []
        
        self.buffer = []

        self.num_calls = 0

    async def tokenize(self, *args, **kwargs):
        def task():
            return self.tokenizer.encode(*args, **kwargs)
        return await asyncio.get_event_loop().run_in_executor(None, task)
        return self.tokenizer.tokenize(*args, **kwargs)
    
    async def detokenize(self, *args, **kwargs):
        def task():
            return self.tokenizer.decode(*args, **kwargs)
        return await asyncio.get_event_loop().run_in_executor(None, task)
        # return self.tokenizer.decode(*args, **kwargs)

    def _complete(self, input_ids, temperature=0, logprobs=1):
        assert len(input_ids) == 1, f"openai model only supports batch size of 1 with logit masks, provided {len(input_ids)}."

        kwargs = {}
        if self.logits_mask[-1] is not None:
            kwargs["logit_bias"] = self.logits_mask[-1]

        kwargs = {
            "model": self.model_identifier,
            "prompt": input_ids.tolist(),
            "max_tokens": self.chunk_size,
            "temperature": temperature,
            "logprobs": logprobs,
            "user": "lmql",
            "stream": True,
            **kwargs
        }

        complete_op = lambda: openai_complete_create(**kwargs)
        return CompleteTask(complete_op, None, continuation_type=None).run(retries=3)
        

    async def advance(self):
        sequence_indices_to_extend_on = []
        prompts_to_extend = []
        for i, buffer in enumerate(self.buffer):
            if await buffer.at_end():
                sequence_indices_to_extend_on.append(i)
                prompts_to_extend.append(self.input_ids[i])

        if len(prompts_to_extend) > 0:
            res = self._complete(np.stack(prompts_to_extend, axis=0))
            response_buffer = OpenAIModelOutputBuffer(res, len(prompts_to_extend), self.tokenizer)
            for i, seq_idx in enumerate(sequence_indices_to_extend_on):
                self.buffer[seq_idx] = response_buffer.buffer(i)

        # advance all self.input_ids by one token from buffer
        next_tokens = []
        for i, buffer in enumerate(self.buffer):
            assert not await buffer.at_end(), f"openai model did not complete decoder head {i} further."
            token, logprob = await buffer.pop()
            next_tokens.append(token)
            self.next_token_scores[i].append(logprob)

        next_tokens = np.array(next_tokens, dtype=np.int64)
        self.input_ids = np.concatenate([self.input_ids, next_tokens.unsqueeze(1)], axis=1)

    def make_api_logits_mask(self, logits, invert):
        if invert:
            masked = (logits >= 0)
        else:
            masked = (logits < 0)
        mask_value = 100 if invert else -100
        return {int(idx): mask_value for idx in np.nonzero(masked)[0]}
    
    async def tokenize(self, text):
        return self.tokenizer(text)["input_ids"]

    async def __aenter__(self):
        self.previous_context_model = OptimisticChunkBasedOpenAIModel.context_model

        bos_token_id = (await self.tokenize("<BOS>"))[0]
        setattr(self, "bos_token_id", bos_token_id)
        eos_token_id = (await self.tokenize("<EOS>"))[0]
        setattr(self, "eos_token_id", eos_token_id)

        OptimisticChunkBasedOpenAIModel.context_model = self
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        OptimisticChunkBasedOpenAIModel.context_model = self.previous_context_model
        self.previous_context_model = None

OptimisticChunkBasedOpenAIModel.context_model = None

class HFModelStatsAdapter:
    @property
    def consumed_tokens(self):
        return 0
    @property
    def num_queries(self):
        return 0
    @property
    def num_generate_calls(self):
        return openai.AsyncConfiguration.get_stats().sum_batch_size
    @property
    def billable_tokens(self):
        return openai.AsyncConfiguration.get_stats().tokens
    
    def reset_stats(self):
        openai.AsyncConfiguration.get_stats().reset()

    def cost_estimate(self, model):
        return openai.AsyncConfiguration.get_stats().cost_estimate(model)

def openai_model(model_identifier):
    # make sure openai org and secret are available
    import lmql.runtime.openai_secret
    
    class ServedModelInterfaceCls:
        def __init__(self) -> None:
            self.model_identifier = model_identifier
            self.decoder_args = None

            self.use_dclib_decoders = False
            self.active_prompting = False

            self.served_model = None
            self._tokenizer = None

            self.bos_token_id = self.get_tokenizer().bos_token_id
            self.eos_token_id = self.get_tokenizer().eos_token_id

            self.adapter = QueryDcLibAdapter(self.get_tokenizer().vocab_size, self.tokenize, self.detokenize, self.bos_token_id, self.eos_token_id)
            self.hf_stats = HFModelStatsAdapter()

        def get_tokenizer(self):
            if self._tokenizer is None:
                self._tokenizer = load_tokenizer("gpt2")
            self.served_model = OptimisticChunkBasedOpenAIModel(model_identifier, self._tokenizer)
            return self._tokenizer

        def set_decoder(self, method, **kwargs):
            # defaults
            DEFAULTS = {
                "max_len": 512,
                "decoder": method
            }
            self.decoder_args = DEFAULTS.copy()

            # custom user arguments
            protected_args = set(["input_ids", "additional_logits_processors", "bos_token_id", "eos_token_id"])
            for key, value in kwargs.items():
                if key in protected_args:
                    print("warning: cannot override runtime determined decoder argument {}.".format(key))
                    continue
                self.decoder_args[key] = value
        
        def get_dclib_model(self):
            bos_token_id = self.get_tokenizer().bos_token_id
            eos_token_id = self.get_tokenizer().eos_token_id

            return DclibOpenAiModel(self.served_model, bos_token_id, eos_token_id, **self.decoder_args)

        async def score_distribution_values(self, prompt, values):
            return await self.adapter.score_distribution_values(prompt, values, self.get_dclib_model())

        async def query(self, prompt, mask_logits_processor, head_input_id_rewriter, active_prompt_rewriter):
            assert self.decoder_args is not None, "Cannot query() a model without calling set_decoder first."

            dclib_model = self.get_dclib_model()
            return await self.adapter.query(prompt, mask_logits_processor, head_input_id_rewriter, active_prompt_rewriter, dclib_model, self.decoder_args)

        async def tokenize(self, text):
            return self.get_tokenizer()(text)["input_ids"]
        
        async def detokenize(self, input_ids):
            return self.get_tokenizer().decode(input_ids)

        def sync_tokenize(self, text):
            return self.get_tokenizer()(text)["input_ids"]
    return ServedModelInterfaceCls


from typing import List, Any, Union, Optional

import asyncio
import numpy as np
from lmql.utils import nputil

from dataclasses import dataclass

from .dclib_global import get_tokenizer
from .dclib_array import DataArray, Continuation, topk, alpha_length_normalized, alpha_length_normalized_det

detokenize_seqs = True

@dataclass
class DecoderGraphSnapshot:
    node_hashes = None
    step_not_updated_count = None

class DecoderGraph:
    def __init__(self):
        self.nodes = {}
        self.node_ids = {}
        self.edges = []

        self.ctr = 0

        self.json_snapshot: DecoderGraphSnapshot = DecoderGraphSnapshot()

    def add_node(self, node):
        uid = f"n{self.ctr}"
        self.ctr += 1
        self.node_ids[node] = uid
        self.nodes[uid] = node
        return uid
    
    def set_pool(self, node, pool_name):
        if not self.node_ids[node] in self.nodes:
            self.add_node(node)
        self.nodes[self.node_ids[node]]["pool"] = pool_name
    
    def add_edge(self, from_node, to_node):
        self.edges.append((self.node_ids[from_node], self.node_ids[to_node]))

    async def json(self, diff: bool = False):
        nodes = []
        
        if self.json_snapshot.node_hashes is None:
            self.json_snapshot.node_hashes = {}
        if self.json_snapshot.step_not_updated_count is None:
            self.json_snapshot.step_not_updated_count = {}
        
        for k, v in self.nodes.items():
            hash = None
            if diff and k in self.json_snapshot.node_hashes:
                # this will ignore changes to nodes that have not been updated for 3 steps (may miss changes to nodes that 
                # are not updated for a long time, careful for now)
                if k in self.json_snapshot.step_not_updated_count and self.json_snapshot.step_not_updated_count[k] > 2:
                    continue
                hash = await v.json_hash()
                if self.json_snapshot.node_hashes[k] == str(hash):
                    self.json_snapshot.step_not_updated_count[k] = self.json_snapshot.step_not_updated_count.get(k, 0) + 1
                    continue
                    
                # else:
                #     print("node changed", k, self.json_snapshot.node_hashes[k], hash)
            if hash is None:
                hash = await v.json_hash()

            nodes.append({
                "id": k,
                **await v.json()
            })
            
            if diff:
                self.json_snapshot.node_hashes[k] = str(hash)
                self.json_snapshot.step_not_updated_count[k] = 0

        return {
            "nodes": nodes,
            "edges": self.edges
        }

class DecoderSequence:
    def __init__(self, input_ids_or_str, logprobs=None, deterministic=None, stop_phrase=None, predecessor=None, user_data=None, sticky_user_data_keys=None, epsilon_node=False):
        assert all([p > DecoderSequence.truncation_threshold for p in logprobs]) if logprobs is not None else True

        if type(input_ids_or_str) == str:
            assert False, "constructing dc.seq() directly from string is not supported anymore"
        elif type(input_ids_or_str) == list:
            input_ids = np.array(input_ids_or_str)
        else:
            input_ids = input_ids_or_str
        
        assert nputil.is_array(input_ids), "input_ids must be a list or numpy array"
        self.input_ids = input_ids
        self.predecessor = predecessor
        
        # if no logprobs are provided assume logprob of 0 (prob of 1)
        if logprobs is None:
            self.logprobs = np.array([0 for _ in range(len(input_ids))])
        else: self.logprobs = logprobs
        assert len(self.logprobs) == len(self.input_ids), "Length of logprobs does not match length of inputs"

        if logprobs is None:
            self.prompt_len = len(input_ids)
        else:
            assert predecessor is not None, "Predecessor is None"
            self.prompt_len = self.predecessor.prompt_len

        self.epsilon_node = epsilon_node

        # if no deterministic is provided assume all tokens to be deterministic
        if deterministic is None: self.deterministic = np.array([True for _ in range(len(input_ids))])
        else: self.deterministic = deterministic
        assert len(self.deterministic) == len(self.input_ids), "Length of determinism status does not match length of inputs"

        # if no stop_phrase is provided assume no tokens to be stop_phrase
        if stop_phrase is None: self.stop_phrase = np.array([False for _ in range(len(input_ids))])
        else: self.stop_phrase = stop_phrase
        assert len(self.stop_phrase) == len(self.input_ids), "Length of determinism status does not match length of inputs"

        # cache for logits when model is called with this sequence
        self.logits = None
        self.raw_logits = None

        if DecoderSequence.graph is not None:
            self.pool = None

        self.user_data = user_data
        self.sticky_user_data_keys = sticky_user_data_keys if sticky_user_data_keys is not None else set()
        self.sticky_user_data_keys.add("head.variable")

        # alternative operation that may called to obtain a score in self.score_next_token if self.raw_logits are not available
        self.score_next_tokens_op = None

        # indicates to dc.rewrite whether this sequence can be rewritten
        self.needs_rewrite = True

    def __hash__(self) -> int:
        return hash(self.id)

    @property
    def id(self):
        if self._id is None: 
            return "<root>"
        return self._id

    def __new__(cls, *args, **kwargs):
        s = super().__new__(cls)

        setattr(s, "_id", f"s_{DecoderSequence.seq_ctr}")

        if DecoderSequence.graph is not None:
            DecoderSequence.graph.add_node(s)

            predecessor = kwargs.get("predecessor")
            if predecessor is not None:
                DecoderSequence.graph.add_edge(predecessor, s)
        
        DecoderSequence.seq_ctr += 1

        return s

    def data(self, key=None, value=None, sticky=False):
        if sticky:
            assert value is not None, "stickiness can only be set when setting a value s.data(key,value,sticky=True)"
        if value is not None:
            if self.user_data is None: self.user_data = {}
            set_path(self.user_data, key, value)
            self.sticky_user_data_keys.add(key)
            return value
        if key is None:
            return self.user_data
        return resolve_path(self.user_data, key)

    @property
    def is_query_constrained(self):
        """All regular sequences are query constrained"""
        return True

    @property
    def num_constraints(self):
        num_const = self.data('head.num_variables_decoded')
        return 0 if num_const is None else num_const

    @property
    def num_det_tokens(self):
        return self.deterministic.sum() - self.prompt_len

    @property
    def num_nondet_tokens(self):
        return (~self.deterministic).sum()

    @property
    def num_tokens(self):
        return len(self.input_ids) - self.prompt_len

    async def detokenized(self, name):
        async def seqtext_provider():
            return str([await get_tokenizer().decode(self.input_ids[1:])][0])
        async def text_provider():
            return str([await get_tokenizer().decode(self.input_ids[-1:])])[2:-2]
        providers = {
            "seqtext": seqtext_provider,
            "text": text_provider
        }
        
        if not hasattr(self, "tokenizer_cache"):
            self.tokenizer_cache = {}
        if name not in self.tokenizer_cache:
            assert name in providers, f"Unknown tokenizer cache provider {name}"
            self.tokenizer_cache[name] = await providers[name]()
        
        return self.tokenizer_cache[name]

    async def json_hash(self):
        o = await self.json()
        if type(o["user_data"]) is dict and "openai-continuations" in o["user_data"]:
            o["user_data"].pop("openai-continuations")
        return hash(str(o))

    async def json(self):
        seqtext = await self.detokenized("seqtext")
        text = await self.detokenized("text")
        if self.epsilon_node: text = [""]
        root = False
        if self.predecessor is None:
            text = await self.detokenized("seqtext")
            # text = str([await get_tokenizer().decode(self.input_ids)])[2:-2],
            root = True

        return {
            "seq_id": self.id,
            # "input_ids": self.input_ids.tolist(),
            "text": [text],
            "seqtext": seqtext,
            "root": root,
            "logprob": self.logprobs[-1],
            "logprobs": self.logprobs[-5:].tolist(),
            "logprobs_det": self.logprobs[-5:][self.deterministic[-5:]].tolist(),
            "logprobs_norm": self.logprobs[-5:][~self.deterministic[-5:]].tolist(),
            "seqlogprob": self.logprobs.sum(),
            **({"deterministic": True} if self.epsilon_node else {}),
            # "score_det": alpha_length_normalized.score(self.logprobs[self.prompt_len:][self.deterministic[self.prompt_len:]], alpha=self.data("scorer_alpha")),
            # "score_nor": alpha_length_normalized.score(self.logprobs[self.prompt_len:][~self.deterministic[self.prompt_len:]], alpha=self.data("scorer_alpha")),
            "score_det": alpha_length_normalized.score(
                self.logprobs[self.prompt_len:][self.deterministic[self.prompt_len:]],),
            "score_nor": alpha_length_normalized.score(
                self.logprobs[self.prompt_len:][~self.deterministic[self.prompt_len:]]),
            "score_tot": alpha_length_normalized_det.score(self.logprobs, self),
            "pool": self.pool,
            "user_data": await self.user_data_json(),
            "token_id": self.input_ids[-1],
            "deterministic_5": self.deterministic[-5:].tolist(),
            "stop_phrase_5": self.stop_phrase[-5:].tolist(),
            "prompt_len" : self.prompt_len,
            "sticky_user_data_keys": list(self.sticky_user_data_keys)
        }

    async def user_data_json(self):
        import json
        class UserJsonEncoder(json.JSONEncoder):
            async def default(self, o):
                try:
                    if type(o) is dict:
                        return {str(k): await self.default(v) for k, v in o.items()}
                    elif type(o) is list:
                        return [await self.default(v) for v in o]
                    elif type(o) is tuple:
                        return tuple(await self.default(list(o)))
                    return super().default(o)
                except:
                    if hasattr(o, "json") and callable(o.json):
                        return await self.default(await o.json())
                    elif type(o) is int:
                        return o
                    elif type(o) is float:
                        return o
                    elif type(o) is bool:
                        return o
                    else:
                        return str(o)
        return await UserJsonEncoder().default(self.user_data)

    def extend_user_data(self, continuation=None, user_data=None):
        assert continuation is None or user_data is None, f"continuation and user_data are mutually exclusive arguments"
        # prepares the user_data dictionary to use with sucessor sequences
        if continuation is None and user_data is None:
            user_data = {}
        elif continuation is not None:
            user_data = deepcopy(continuation.user_data)
        elif user_data is not None:
            user_data = deepcopy(user_data)
        else:
            user_data = deepcopy(continuation.user_data)
        for sk in self.sticky_user_data_keys:
            if user_data is None:
                user_data = {}
            set_path(user_data, sk, self.data(sk), create_missing=True, replace=False)
        return user_data

    def extend(self, continuation):
        stop_phrase = self.detect_stop_phrase(continuation)
        return DecoderSequence(
            input_ids_or_str=np.concatenate([self.input_ids, continuation.token.reshape(1)]), 
            logprobs=np.concatenate([self.logprobs, continuation.logprob.reshape(1)]),
            # deterministic tokens are only extended in DeterministicDecoderSequence.extend.
            # So here, all extended tokens are non-deterministic, i.e. model predictions.
            deterministic=np.concatenate([self.deterministic, np.array([False])]),
            stop_phrase=stop_phrase,
            predecessor=self,
            user_data=self.extend_user_data(continuation),
            sticky_user_data_keys=self.sticky_user_data_keys
        )

    def detect_stop_phrase(self, continuation):
        old_stop_phrase = self.stop_phrase
        new_stop_phrase = np.concatenate([old_stop_phrase, np.array([False])])
        stop_phrases = self.data("head.stopping_phrases.tokenized")

        if stop_phrases is None:
            return new_stop_phrase

        ids = np.concatenate([self.input_ids, continuation.token.reshape(1)]),
        for stop in stop_phrases:
            len_stop = len(stop)
            if ids[-len_stop:] == stop:
                new_stop_phrase[-len_stop:] = True
                break

        return new_stop_phrase

    def __len__(self):
        return len(self.input_ids)

    def __repr__(self) -> str:
        ids = ", ".join([str(i) for i in self.input_ids[-10:]])
        return f"<seq token_len={len(self.input_ids)} ids=[... {ids}]>"

    async def text(self) -> str:
        return str([await get_tokenizer().decode(self.input_ids)])

    async def str(self, full=False) -> str:
        ids = ", ".join([str(i) for i in self.input_ids[-10:]])
        if detokenize_seqs:
            s = await get_tokenizer().decode(self.input_ids)
            if not full:
                s = "..." + s[-40:]
            return f"<seq token_len={len(self.input_ids)} s={str([s])[1:-1]} ids=[... {ids}]>"
        else:
            return f"<seq token_len={len(self.input_ids)} ids=[... {ids}]>"

    def __str__(self):
        ids = ", ".join([str(i) for i in self.input_ids[-10:]])
        return f"<seq token_len={len(self.input_ids)} ids=[... {ids}]>"

    def is_done(self):
        return self.input_ids[-1] == get_tokenizer().eos_token_id

    def make_successors(self, next_tokens, next_token_scores, logits, user_data=None):
        # remove very low scoring tokens (likely they were masked and therefore score low)
        next_tokens = np.stack([t for t, s in zip(next_tokens, next_token_scores) if s > DecoderSequence.truncation_threshold], axis=0)
        next_token_scores = np.stack([s for s in next_token_scores if s > DecoderSequence.truncation_threshold], axis=0)

        return Continuation(next_tokens, next_token_scores, user_data)
# global counter for all sequences created in this process for identification purposes
DecoderSequence.seq_ctr = 0

DecoderSequence.graph = None
# tokens with a logprob lower than this will be ignored and not expanded during decoding
DecoderSequence.truncation_threshold = -3e38

def resolve_path(d, path):
    if d is None:
        return None
    segments = path.split(".")
    for segment in segments:
        if segment not in d:
            return None
        d = d[segment]
    return d

def set_path(d, path, value, create_missing=False, replace=True):
    if d is None:
        return
    segments = path.split(".")
    for segment in segments[:-1]:
        if segment not in d:
            if create_missing:
                d[segment] = {}
            else:
                return
        d = d[segment]
    if not replace and segments[-1] in d.keys():
        return
    d[segments[-1]] = value

def deepcopy(d):
    # deepcopy dict and iterables
    if type(d) is dict:
        return {k: deepcopy(v) for k, v in d.items()}
    elif type(d) is list:
        return [deepcopy(v) for v in d]
    else:
        return d

def deepmerge(a, b):
    if a is None: return b
    if b is None: return a
    for k,v in b.items():
        if k in a and isinstance(a[k], dict) and isinstance(b[k], dict):
            deepmerge(a[k], b[k])
        else:
            a[k] = b[k]
    return a

class DeterministicDecoderSequence(DecoderSequence):
    def __init__(self, input_ids, logprobs, deterministic, stop_phrase, next_ids, next_logprobs=None, next_deterministic=None, predecessor=None, user_data=None, needs_rewrite=False, sticky_user_data_keys=None):
        super().__init__(input_ids, logprobs, deterministic, stop_phrase, predecessor, user_data=user_data.copy() if user_data is not None else None, sticky_user_data_keys=sticky_user_data_keys)
        self.next_ids = next_ids
        self.next_logprobs = next_logprobs
        self.next_deterministic = next_deterministic

        self.needs_rewrite = needs_rewrite

        if next_logprobs is not None: assert len(next_logprobs) == len(next_ids), "Length of deterministic continuation did not match length of provided logprobs"
        if next_deterministic is not None: assert len(next_deterministic) == len(next_ids), "Length of determinism status did not match length of provided logrprobs"

        self.align_user_data()

    def align_user_data(self):
        # lmql-specific user data should be different for deterministic sequences
        head_variable = resolve_path(self.user_data, "head.variable")
        if head_variable is not None:
            if "before(" in head_variable:
                head_variable = head_variable.split(":before(", 1)[0]
            if head_variable == "__done__":
                set_path(self.user_data, "head.variable", "__done__")
            else:
                if len(self.next_ids) > 0:
                    set_path(self.user_data, "head.variable", head_variable + ":before(" + str(len(self.next_ids)) + ")")
                else:
                    set_path(self.user_data, "head.variable", head_variable)
            # deterministic sequences don't have stopping phrases
            set_path(self.user_data, "head.stopping_phrases", {"tokenized": [], "text": []})
        else:
            set_path(self.user_data, "head.variable", "<prompt>")
        if len(self.next_ids) > 0:
            set_path(self.user_data, "head.mask", "{token_id=" + str(self.next_ids[0]) + "}")
        else:
            set_path(self.user_data, "head.mask", "<not available yet>")
    
    @property
    def is_query_constrained(self):
        """Deterministic sequences are not query constrained, as long as they are fixed to their predetermined content."""
        return True if len(self.next_ids) == 0 else False

    async def json(self):
        seqtext = await self.detokenized("seqtext")
        text = await self.detokenized("text")
        root = False
        if self.predecessor is None:
            text = await self.detokenized("seqtext")
            # text = str([await get_tokenizer().decode(self.input_ids)])[2:-2],
            root = True
        
        return {
            "seq_id": self.id,
            # "input_ids": self.input_ids.tolist(),
            "text": [text],
            "seqtext": seqtext,
            "root": root,
            "logprob": self.logprobs[-1],
            "seqlogprob": self.logprobs.sum(),
            "pool": self.pool,
            "deterministic": True,
            "next_ids": self.next_ids.tolist(),
            "next_logprobs": self.next_logprobs.tolist() if self.next_logprobs is not None else None,
            "user_data": await self.user_data_json(),
            "token_id": self.input_ids[-1],
            "deterministic_5": self.deterministic[-5:].tolist(),
            "stop_phrase_5": self.stop_phrase[-5:].tolist(),
            "score_det": alpha_length_normalized.score(self.logprobs[self.prompt_len:][self.deterministic[self.prompt_len:]]),
            "score_nor": alpha_length_normalized.score(self.logprobs[self.prompt_len:][~self.deterministic[self.prompt_len:]]),
            "score_tot": alpha_length_normalized_det.score(self.logprobs, self),
            **({"done": True} if self.data("head.variable") == "__done__" else {}),
            "sticky_user_data_keys": list(self.sticky_user_data_keys)
        }

    def expand(self):
        """ Returns the sequence that corresponds to the full expansion of this pre-determined deterministic sequence. """
        s = self
        while type(s) is DeterministicDecoderSequence and len(s.next_ids) > 0:
            s = s.extend(Continuation(s.next_ids[0], s.next_logprobs[0], s.user_data))
        return s

    def extend_user_data(self, continuation=None, user_data=None):
        # additionally add self.user_data for deterministic sequence extensions
        return deepmerge(deepcopy(self.user_data), super().extend_user_data(continuation=continuation, user_data=user_data))

    def extend(self, continuation):
        # if not more predetermined tokens are left, just extend as usual and return a regular seq()
        if len(self.next_ids) <= 0:
            return super().extend(continuation)

        assert continuation.token == self.next_ids[0], "deterministic sequences must be extended by the predetermined next token: provided: " + str(continuation.token) + ", predetermined: " + str(self.next_ids[0])

        extended_input_ids = np.concatenate([self.input_ids, continuation.token.reshape(1)])
        extended_logprobs = np.concatenate([self.logprobs, continuation.logprob.reshape(1)])
        extended_deterministic = np.concatenate([self.deterministic, np.array([True]) if self.next_deterministic is None else self.next_deterministic[0:1]])

        reduced_next_ids = self.next_ids[1:]
        reduced_next_logprobs = self.next_logprobs[1:] if self.next_logprobs is not None else None
        reduced_deterministic = None if self.next_deterministic is None else self.next_deterministic[1:]

        user_data = self.extend_user_data(continuation)

        stop_phrase = self.detect_stop_phrase(continuation)
        if self.data("injected_stop_phrase"):
            print(self)
            assert not extended_deterministic[-1]

        return DeterministicDecoderSequence(
            input_ids=extended_input_ids, 
            logprobs=extended_logprobs, 
            deterministic=extended_deterministic,
            stop_phrase=stop_phrase,#np.concatenate([self.stop_phrase, self.data("injected_stop_phrase")]),
            next_ids=reduced_next_ids,
            next_logprobs=reduced_next_logprobs,
            next_deterministic=reduced_deterministic,
            predecessor=self, 
            user_data=user_data,
            needs_rewrite=self.needs_rewrite,
            sticky_user_data_keys=self.sticky_user_data_keys,
        )

    def __repr__(self) -> str:
        ids = ", ".join([str(i) for i in self.input_ids[-10:]])
        if (len(self.next_ids) > 0): 
            ids = "... " + ids
        next_ids = ", ".join([str(i) for i in self.next_ids[:10]])
        if len(self.next_ids) > 10: 
            next_ids = "..." + next_ids
        return f"<detseq token_len={len(self.input_ids)} ids=[{ids}] next_ids=[{next_ids}]>"

    async def str(self, full=False) -> str:
        ids = ", ".join([str(i) for i in self.input_ids[-10:]])
        if (len(self.next_ids) > 0): 
            ids = "... " + ids
        next_ids = ", ".join([str(i) for i in self.next_ids[:10]])
        if len(self.next_ids) > 10: 
            next_ids = "..." + next_ids
        if detokenize_seqs:
            s = await get_tokenizer().decode(self.input_ids)
            if not full:
                s = "..." + s[-40:]
            return f"<detseq token_len={len(self.input_ids)} s={str([s])[1:-1]} ids=[{ids}] next_ids=[{next_ids}]>"
        else:
            return f"<detseq token_len={len(self.input_ids)} ids=[{ids}] next_ids=[{next_ids}]>"

    def make_successors(self, next_tokens, next_token_scores, logits, user_data=None):
        # if no more predetermined tokens are left, just make successors as usual
        if len(self.next_ids) <= 0:
            return super().make_successors(next_tokens, next_token_scores, logits, user_data=user_data)

        # logits here are unprocessed
        raw_logits = logits
        # TODO: using the raw token score will not consider sampling temperature (but it should?)
        # e.g. deterministic tokens are scored according to their raw logprob, not their logprob after temperature is applied
        # ignore next_tokens and next_token_scores, only produce predetermined next_ids as only successors
        predetermined_token = self.next_ids[0]
        
        # obtain score (either precomputed or via the passed logits)
        if self.next_logprobs is not None:
            score = self.next_logprobs[0]
        else:
            score = raw_logits[predetermined_token]
            if score < DecoderSequence.truncation_threshold:
                print("warning: a deterministic token scored below the truncation threshold ({})".format(DecoderSequence.truncation_threshold))
        
        return Continuation(predetermined_token, score, user_data)

def is_deterministic(s):
    return issubclass(type(s), DeterministicDecoderSequence)

def next_is_deterministic(s):
    return is_deterministic(s) and len(s.next_ids) > 0

def is_lmql_valid(s):
    return s.data("head.valid") != False or s.data("head.final") != "fin"

def set_record_graph():
    DecoderSequence.graph = DecoderGraph()

def set_truncation_threshold(threshold):
    DecoderSequence.truncation_threshold = threshold

def detseq(
    ids: np.ndarray,
    logprobs: np.ndarray,
    deterministic: np.ndarray,
    stop_phrase: np.ndarray,
    next_ids: List[int],
    next_logprobs: List[float] = None,
    next_deterministic: Optional[np.ndarray] = None,
    predecessor=None, 
    user_data=None,
    needs_rewrite=True,
    sticky_user_data_keys=None):
    
    return DeterministicDecoderSequence(
        input_ids=ids, 
        logprobs=logprobs, 
        deterministic=deterministic,
        stop_phrase=stop_phrase,
        next_ids=next_ids,
        next_logprobs=next_logprobs,
        next_deterministic=next_deterministic,
        predecessor=predecessor, 
        user_data=user_data,
        needs_rewrite=needs_rewrite,
        sticky_user_data_keys=sticky_user_data_keys,
    )

def seq(ids: List[int], logprobs:Optional[np.ndarray]=None, deterministic:Optional[np.ndarray]=None):
    return DecoderSequence(ids, logprobs=logprobs, deterministic=deterministic)

def named(seqs: List[DecoderSequence] = None, name=None):
    for s in seqs:
        if hasattr(s, "pool"):
            s.pool = name
    return seqs

class FinishException(Exception):
    def __init__(self, result_sequences: List[DecoderSequence]):
        super().__init__()
        self.result_sequences = result_sequences

def finish(ar: Union[DataArray, List[DecoderSequence], DecoderSequence], **kwargs):
    """
    Terminates the decoding process and returns the provided (array of) sequences as the final
    result sequences in order of the underlying array/list. 1
    """
    if type(ar) is DecoderSequence:
        raise FinishException([ar])
    elif type(ar) is list:
        raise FinishException(ar)
    else:
        raise FinishException(list(ar.flatten().sequences.values())[0])

def token_unique(ar: DataArray, prefer=None, flatten=False):
    """
    Like unique() but compares tokens instead of sequence identities.
    """
    if prefer is None:
        prefer = lambda s,t: True

    if flatten:
        original_shape = ar.shape
        ar = ar.flatten()

    def op_unique(sqs):
        assert type(sqs) is list, f"token_unique expects a DataArray of lists of DecoderSequences, but got {type(sqs)} as data item."
        assert len(sqs) == 0 or issubclass(type(sqs[0]), DecoderSequence), f"token_unique expects a DataArray of lists of DecoderSequences, but got a list of {type(sqs[0])} as data item."

        by_token_ids = {}
        for s in sqs:
            token_ids = tuple(s.input_ids.tolist())
            curr = by_token_ids.get(token_ids, None)
            if curr is None or prefer(s, curr):
                by_token_ids[token_ids] = s
        return list(by_token_ids.values())

    res = ar.element_wise(op_unique)

    if flatten:
        return res.reshape(original_shape)
    else:
        return res


def repeat(ar, n, scorer=None):
    """
    Repeats the sequences in `ar` until a total of `n` sequences are in the array. 
    
    If no integer-valued multiple of ar's contents fit, only the top n sequences are repeated in the last tile.
    """
    if len(ar) == n: return ar
    elif len(ar) > n: 
        return topk(ar, n, scorer=scorer)
    else: 
        return ar + repeat(ar, n - len(ar), scorer=scorer)



from dataclasses import dataclass, field
import tokenize
from io import StringIO
from typing import Any, List
import astunparse
import ast
import sys
import termcolor

import lmql.runtime.dclib.decoders
from lmql.runtime.dclib import get_all_decoders

class FragmentParserError(Exception): pass

@dataclass
class LMQLDecoderConfiguration:
    method: ast.AST
    decoding_args: List[ast.keyword]

@dataclass
class LMQLDistributionClause:
    variable_name: str
    values: List[str]

@dataclass
class LMQLQuery:
    # all code that comes before the first decode keyword
    prologue: list = field(default_factory=list)

    decode_str: list = field(default_factory=list)
    decode = None
    
    prompt_str: list = field(default_factory=list)
    prompt = None

    from_str: list = field(default_factory=list)
    from_ast = None

    where_str: list = field(default_factory=list)
    where = None
    where_expr = None

    distribution_str: list = field(default_factory=list)
    distribution = None
    distribution_expr = None

    scoring_str: list = field(default_factory=list)
    scoring = None

    scope = None

def is_keyword(tok, kw):
    return tok.type == tokenize.NAME and tok.string.lower() == kw.lower()

def remove_indentation(s, oneline=False):
    # print([s])
    lines = []

    min_indent = " " * len(s)

    for line in s.split("\n"):
        if line.strip() == "" or line == "\\": continue
        unindented_line = line.lstrip()
        line_indent = len(line) - len(unindented_line)
        min_indent = min_indent if len(min_indent) < line_indent else line[:line_indent]
        # print(unindented_line, [line[:line_indent], line], line_indent)
        
        # buf = StringIO(line)
        # toks = [t for t in tokenize.generate_tokens(buf.readline)]
        # print(tokenize.tok_name[toks[-4].type])
        
        # lines.append(line.lstrip())

    # print("min_indent is", len(min_indent), [min_indent])

    for line in s.split("\n"):
        if line.strip() == "" or line == "\\": continue
        assert line.startswith(min_indent)

        lines.append(line[len(min_indent):])

    if oneline:
        return " \\\n".join(lines)
    else:
        return "\n".join(lines)

def remove_comments(s):
    if type(s) is str: return s
    assert type(s) is list
    return tokenize.untokenize([transform_token(t) for t in s if t.type != tokenize.COMMENT])

def transform_token(t):
    if is_keyword(t, "as"):
        # return tokenize.Token
        # print(t)
        return tokenize.TokenInfo(tokenize.OP, "=", start = t.start, end = t.end, line = t.line)

    return t

def ast_parse(s, unindent=False, oneline=False, loc=None):
    try:
        s = remove_comments(s)
        if unindent: s = remove_indentation(s, oneline=oneline)
        return ast.parse(s)
    except SyntaxError as e:
        msg = ""
        
        msg += "Failed to parse {} clause of the query ({}):\n\n".format(loc, e.msg)
        
        for lineno, line in enumerate(s.split("\n")):
            if lineno + 1 == e.lineno:
                msg += "\t" + line[:e.offset - 1]
                msg += termcolor.colored(line[e.offset - 1:], "red") + "\n"
                msg += "\t" + ((e.offset - 1) * " ") + "^\n"
                break
            elif abs(lineno + 1 - e.lineno) < 2:
                msg += "\t" + line + "\n"
        
        raise FragmentParserError(msg)

def tok_str(tok):
    if tok.type == tokenize.NAME and tok.string == "AND":
        return "and"
    if tok.type == tokenize.NAME and tok.string == "OR":
        return "and"
    if tok.type == tokenize.NAME and tok.string == "IN":
        return "in"
    if tok.type == tokenize.NAME and tok.string == "NOT":
        return "not"
    if tok.type == tokenize.NAME and tok.string == "AS":
        return "as"
    return tok.string

class LanguageFragmentParser:
    def __init__(self):
        self.state = "start" # "decode" | "prompt" | "where" | "scoring" | "import"
        self.query = LMQLQuery()
        self.paren_count = 0

    def parse(self, readline):
        for i, tok in enumerate(tokenize.generate_tokens(readline)):
            self.digest(tok)

        # print("decode_str", remove_comments(self.query.decode_str))
        # print("prompt_str", remove_comments(self.query.prompt_str))
        # print("from_str", remove_comments(self.query.from_str))

        self.prologue_transform()
        self.ast_parse()
        self.syntax_validation()
        self.ast_transform()

        return self.query

    def prologue_transform(self):
        # translate prologue tokens into str
        self.query.prologue = tokenize.untokenize(self.query.prologue)

    def ast_transform(self):
        if self.query.distribution is not None:
            # this structure is already validated (see self.syntax_validation)
            variable_name = self.query.distribution.left.id
            self.query.distribution = LMQLDistributionClause(variable_name, self.query.distribution.comparators[0])

    def syntax_validation(self):
        # validate decode clause
        if len(self.query.decode_str) == 0:
            raise FragmentParserError("A query must start with a decode clause (e.g. ARGMAX).")
        
        # validate from clause
        if len(self.query.from_str) == 0:
            raise FragmentParserError("The query is missing a FROM clause (e.g. FROM 'gpt-2').")

        # validation distribution clause
        if self.query.distribution is not None:
            error_msg = "The distribution clause must be formed like 'VAR in [list of values]'."
            if type(self.query.distribution) is not ast.Compare:
                raise FragmentParserError(error_msg)
            if type(self.query.distribution.ops[0]) is not ast.In:
                raise FragmentParserError(error_msg)
            variable_node = self.query.distribution.left
            if type(variable_node) is not ast.Name:
                raise FragmentParserError(error_msg)
            values_list = self.query.distribution.comparators[0]

    def ast_parse(self):
        # parse decode, prompt and from
        self.query.decode = ast_parse(self.query.decode_str, loc="decode").body[0].value
        self.query.prompt = ast_parse(self.query.prompt_str, unindent=True, loc="prompt").body
        self.query.from_ast = ast_parse(self.query.from_str, unindent=True, loc="from").body[0]

        where_body = ast_parse(self.query.where_str, unindent=True, oneline=True, loc="where").body
        if len(where_body) > 0:
            self.query.where = where_body[0]
        else:
            self.query.where = None
        
        # parse distribution clause if present
        self.query.distribution = ast_parse(self.query.distribution_str, unindent=True, loc="distribution").body
        if len(self.query.distribution) > 0: 
            self.query.distribution = self.query.distribution[0].value
        else:
            self.query.distribution = None

        scoring_body = ast_parse(self.query.scoring_str, unindent=True, loc="scoring").body
        if len(scoring_body) > 0:
            self.query.scoring = scoring_body[0]
        else:
            self.query.scoring = None

    def digest(self, tok):
        if self.state == "start":
            if tok.type == tokenize.NAME:
                # when we encounter the first decoder keyword, we switch to the query parsing state
                if tok.string.lower() in get_all_decoders():
                    self.query.decode_str += [tok]
                    self.state = "decode"
                    return
            
            # otherwise add token to prologue tokens (e.g. imports, comments, function definitions)
            self.query.prologue.append(tok)
            return
        elif self.state == "decode":
            if tok.type == tokenize.STRING and self.paren_count == 0:
                self.state = "prompt"
                self.query.prompt_str += [tok]
                return
            
            self.query.decode_str += [tok]

            if tok.type == tokenize.OP and tok_str(tok) == "(":
                self.paren_count += 1
            elif tok.type == tokenize.OP and tok_str(tok) == ")":
                self.paren_count -= 1
            if self.paren_count == 0:
                self.state = "prompt"
        elif self.state == "prompt":
            if is_keyword(tok, "FROM"):
                self.state = "from"
                return
            if is_keyword(tok, "WHERE"):
                self.state = "where"
                return
            if is_keyword(tok, "SCORING"):
                self.state = "scoring"
                return
            if is_keyword(tok, "DISTRIBUTION"):
                self.state = "distribution"
                return
            
            # if last token is NAME and current is str
            if len(self.query.prompt_str) > 0 and self.query.prompt_str[-1].type == tokenize.NAME and \
                tok.type == tokenize.STRING:
                last_tok = self.query.prompt_str[-1]
                try:
                    untokenized = tokenize.untokenize([last_tok, tok]).split(last_tok.string)[1]
                    if not untokenized.startswith(" "):
                        # print("merge name and tok", last_tok, tok)
                        self.query.prompt_str.pop(-1)
                except:
                    pass
            self.query.prompt_str += [tok]
        elif self.state == "from":
            if is_keyword(tok, "WHERE"):
                self.state = "where"
                return
            if is_keyword(tok, "SCORING"):
                self.state = "scoring"
                return
            if is_keyword(tok, "DISTRIBUTION"):
                self.state = "distribution"
                return

            self.query.from_str += [tok]
        elif self.state == "where":
            if is_keyword(tok, "SCORING"):
                self.state = "scoring"
                return
            if is_keyword(tok, "DISTRIBUTION"):
                self.state = "distribution"
                return

            self.query.where_str += [tok]
        elif self.state == "distribution":
            self.query.distribution_str += [tok]



import numpy as np

def is_array(x):
    return isinstance(x, np.ndarray)

def item(a):
    if is_array(a): 
        assert len(a) == 1 and a.ndim == 1
        return a[0]
    else: return a

def ensure_array(v, dtype=None):
    if v is None: return None
    if is_array(v): return v
    else: return np.array(v, dtype=dtype)


def log_softmax(a):
    return a - np.log(np.sum(np.exp(a)))

def topk(a, k:int, sorted: bool = False, axis=-1):
    assert k > 0, "topk(): k must be > 0"
    idx = np.argpartition(a, -k, axis=axis)[..., -k:]
    if sorted: 
        index_values = np.take_along_axis(a, idx, axis=axis)
        idx = np.take_along_axis(idx, np.argsort(index_values, axis=axis), axis=axis)
        # reverse order of indices 
        idx = idx[..., ::-1]
    values = np.take_along_axis(a, idx, axis=axis)
    return values, idx

def multinomial(logprobs, num_samples=1):
    if num_samples == 0:
        return np.array([], dtype=np.int64), np.array([], dtype=np.float32)
    # renormalize logprobs
    logprobs = logprobs - np.log(np.exp(logprobs).sum())
    probs = np.exp(logprobs)

    num_samples_i = min(num_samples, len(logprobs), (logprobs > -np.inf).sum())
    next_token_id = np.random.choice(len(logprobs), p=probs, size=num_samples_i, replace=False)
    next_token_score = logprobs[next_token_id]

    return next_token_id, next_token_score

def unsqueeze(a, axis):
    if is_array(a):
        return np.expand_dims(a, axis)
    else:
        return a